import concurrent.futures
import copy
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Optional

from sqlalchemy import text

from core.base import (
    DatabaseConfig,
    VectorDBProvider,
    VectorEntry,
    VectorSearchResult,
)
from core.base.abstractions import VectorSearchSettings

from .vecs import (
    Client,
    Collection,
    IndexArgsHNSW,
    IndexMeasure,
    IndexMethod,
    create_client,
)

logger = logging.getLogger(__name__)


class PostgresVectorDBProvider(VectorDBProvider):
    def __init__(self, config: DatabaseConfig, *args, **kwargs):
        super().__init__(config)
        self.collection: Optional[Collection] = None
        connection_string = kwargs.get("connection_string", None)
        if not connection_string:
            raise ValueError(
                "Please provide a valid `connection_string` to the `PostgresVectorDBProvider`."
            )
        self.vx: Client = create_client(connection_string=connection_string)
        if not self.vx:
            raise ValueError(
                "Error occurred while attempting to connect to the pgvector provider."
            )
        self.project_name = kwargs.get("project_name", None)
        if not self.project_name:
            raise ValueError(
                "Please provide a valid `project_name` to the `PostgresVectorDBProvider`."
            )
        dimension = kwargs.get("dimension", None)
        if not dimension:
            raise ValueError(
                "Please provide a valid `dimension` to the `PostgresVectorDBProvider`."
            )

        self._initialize_vector_db(dimension)
        logger.info(
            f"Successfully initialized PGVectorDB for project: {self.project_name}"
        )

    def _initialize_vector_db(self, dimension: int) -> None:
        # Create extension for trigram similarity
        with self.vx.Session() as sess:
            sess.execute(text("CREATE EXTENSION IF NOT EXISTS pg_trgm;"))
            sess.execute(text("CREATE EXTENSION IF NOT EXISTS btree_gin;"))
            sess.commit()

        self.collection = self.vx.get_or_create_vector_table(
            name=self.project_name, dimension=dimension
        )
        self.create_index()

    def upsert(self, entry: VectorEntry) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert`."
            )

        self.collection.upsert(
            records=[
                (
                    entry.fragment_id,
                    entry.extraction_id,
                    entry.document_id,
                    entry.user_id,
                    entry.collection_ids,
                    entry.vector.data,
                    entry.text,
                    entry.metadata,
                )
            ]
        )

    def upsert_entries(self, entries: list[VectorEntry]) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert_entries`."
            )

        self.collection.upsert(
            records=[
                (
                    entry.fragment_id,
                    entry.extraction_id,
                    entry.document_id,
                    entry.user_id,
                    entry.collection_ids,
                    entry.vector.data,
                    entry.text,
                    entry.metadata,
                )
                for entry in entries
            ]
        )

    def semantic_search(
        self, query_vector: list[float], search_settings: VectorSearchSettings
    ) -> list[VectorSearchResult]:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `semantic_search`."
            )
        results = self.collection.semantic_search(
            vector=query_vector, search_settings=search_settings
        )
        return [
            VectorSearchResult(
                fragment_id=result[0],  # type: ignore
                extraction_id=result[1],  # type: ignore
                document_id=result[2],  # type: ignore
                user_id=result[3],  # type: ignore
                collection_ids=result[4],  # type: ignore
                text=result[5],  # type: ignore
                score=1 - float(result[6]),  # type: ignore
                metadata=result[7],  # type: ignore
            )
            for result in results
        ]

    def full_text_search(
        self, query_text: str, search_settings: VectorSearchSettings
    ) -> list[VectorSearchResult]:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `full_text_search`."
            )
        return self.collection.full_text_search(
            query_text=query_text, search_settings=search_settings
        )

    def hybrid_search(
        self,
        query_text: str,
        query_vector: list[float],
        search_settings: VectorSearchSettings,
        *args,
        **kwargs,
    ) -> list[VectorSearchResult]:
        if search_settings.hybrid_search_settings is None:
            raise ValueError(
                "Please provide a valid `hybrid_search_settings` in the `search_settings`."
            )
        if (
            search_settings.hybrid_search_settings.full_text_limit
            < search_settings.search_limit
        ):
            raise ValueError(
                "The `full_text_limit` must be greater than or equal to the `search_limit`."
            )

        semantic_settings = copy.deepcopy(search_settings)
        semantic_settings.search_limit += search_settings.offset

        full_text_settings = copy.deepcopy(search_settings)
        full_text_settings.hybrid_search_settings.full_text_limit += (  # type: ignore
            search_settings.offset
        )

        # Use ThreadPoolExecutor to run searches in parallel
        with ThreadPoolExecutor(max_workers=2) as executor:
            semantic_future = executor.submit(
                self.semantic_search, query_vector, semantic_settings
            )
            full_text_future = executor.submit(
                self.full_text_search, query_text, full_text_settings
            )

            # Wait for both searches to complete
            concurrent.futures.wait([semantic_future, full_text_future])

        semantic_results: list[VectorSearchResult] = semantic_future.result()
        full_text_results: list[VectorSearchResult] = full_text_future.result()

        semantic_limit = search_settings.search_limit
        full_text_limit = (
            search_settings.hybrid_search_settings.full_text_limit
        )
        semantic_weight = (
            search_settings.hybrid_search_settings.semantic_weight
        )
        full_text_weight = (
            search_settings.hybrid_search_settings.full_text_weight
        )
        rrf_k = search_settings.hybrid_search_settings.rrf_k

        # Combine results using RRF
        combined_results = {
            result.fragment_id: {
                "semantic_rank": rank,
                "full_text_rank": full_text_limit,
                "data": result,
            }
            for rank, result in enumerate(semantic_results, 1)
        }

        semantic_limit = search_settings.search_limit
        full_text_limit = (
            search_settings.hybrid_search_settings.full_text_limit
        )
        semantic_weight = (
            search_settings.hybrid_search_settings.semantic_weight
        )
        full_text_weight = (
            search_settings.hybrid_search_settings.full_text_weight
        )
        rrf_k = search_settings.hybrid_search_settings.rrf_k
        # Combine results using RRF
        combined_results = {
            result.fragment_id: {
                "semantic_rank": rank,
                "full_text_rank": full_text_limit,
                "data": result,
            }
            for rank, result in enumerate(semantic_results, 1)
        }

        for rank, result in enumerate(full_text_results, 1):
            if result.fragment_id in combined_results:
                combined_results[result.fragment_id]["full_text_rank"] = rank
            else:
                combined_results[result.fragment_id] = {
                    "semantic_rank": semantic_limit,
                    "full_text_rank": rank,
                    "data": result,
                }

        # Filter out non-overlapping results
        combined_results = {
            k: v
            for k, v in combined_results.items()
            if v["semantic_rank"] <= semantic_limit * 2  # type: ignore
            and v["full_text_rank"] <= full_text_limit * 2  # type: ignore
        }

        # Calculate RRF scores
        for result in combined_results.values():  # type: ignore
            semantic_score = 1 / (rrf_k + result["semantic_rank"])  # type: ignore
            full_text_score = 1 / (rrf_k + result["full_text_rank"])  # type: ignore
            result["rrf_score"] = (  # type: ignore
                semantic_score * semantic_weight
                + full_text_score * full_text_weight
            ) / (semantic_weight + full_text_weight)

        # Sort by RRF score and apply offset and limit
        sorted_results = sorted(
            combined_results.values(),
            key=lambda x: x["rrf_score"],  # type: ignore
            reverse=True,
        )
        offset_results = sorted_results[
            search_settings.offset : search_settings.offset
            + search_settings.search_limit
        ]

        return [
            VectorSearchResult(
                fragment_id=result["data"].fragment_id,  # type: ignore
                extraction_id=result["data"].extraction_id,  # type: ignore
                document_id=result["data"].document_id,  # type: ignore
                user_id=result["data"].user_id,  # type: ignore
                collection_ids=result["data"].collection_ids,  # type: ignore
                text=result["data"].text,  # type: ignore
                score=result["rrf_score"],  # type: ignore
                metadata={
                    **result["data"].metadata,  # type: ignore
                    "semantic_rank": result["semantic_rank"],
                    "full_text_rank": result["full_text_rank"],
                },
            )
            for result in offset_results
        ]

    def create_index(
        self,
        index_type=IndexMethod.hnsw,
        measure=IndexMeasure.cosine_distance,
        index_options=None,
    ):
        if self.collection is None:
            raise ValueError("Collection is not initialized.")

        if index_options is None:
            index_options = IndexArgsHNSW(
                m=16, ef_construction=64
            )  # Default HNSW parameters

        self.collection.create_index(
            method=index_type,
            measure=measure,
            index_arguments=index_options,
            replace=True,
        )

    def delete(
        self,
        filters: dict[str, Any],
    ) -> dict[str, dict[str, str]]:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `delete`."
            )

        return self.collection.delete(filters=filters)

    def assign_document_to_collection(
        self, document_id: str, collection_id: str
    ) -> None:
        """
        Assign a document to a collection in the vector database.

        Args:
            document_id (str): The ID of the document to assign.
            collection_id (str): The ID of the collection to assign the document to.

        Raises:
            ValueError: If the collection is not initialized.
        """
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `assign_document_to_collection`."
            )

        table_name = self.collection.table.name
        query = text(
            f"""
            UPDATE vecs."{table_name}"
            SET collection_ids = array_append(collection_ids, :collection_id)
            WHERE document_id = :document_id AND NOT (:collection_id = ANY(collection_ids))
            RETURNING document_id
            """
        )

        with self.vx.Session() as sess:
            result = sess.execute(
                query,
                {"document_id": document_id, "collection_id": collection_id},
            ).fetchone()
            sess.commit()

        if not result:
            logger.warning(
                f"Document {document_id} not found or already assigned to collection {collection_id}"
            )

    def remove_document_from_collection(
        self, document_id: str, collection_id: str
    ) -> None:
        """
        Remove a document from a collection in the vector database.

        Args:
            document_id (str): The ID of the document to remove.
            collection_id (str): The ID of the collection to remove the document from.

        Raises:
            ValueError: If the collection is not initialized.
        """
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `remove_document_from_collection`."
            )

        table_name = self.collection.table.name
        query = text(
            f"""
            UPDATE vecs."{table_name}"
            SET collection_ids = array_remove(collection_ids, :collection_id)
            WHERE document_id = :document_id AND :collection_id = ANY(collection_ids)
            RETURNING document_id
            """
        )

        with self.vx.Session() as sess:
            result = sess.execute(
                query,
                {"document_id": document_id, "collection_id": collection_id},
            ).fetchone()
            sess.commit()

        if not result:
            logger.warning(
                f"Document {document_id} not found in collection {collection_id} or already removed"
            )

    def remove_collection_from_documents(self, collection_id: str) -> None:
        if self.collection is None:
            raise ValueError("Collection is not initialized.")

        table_name = self.collection.table.name
        query = text(
            f"""
            UPDATE vecs."{table_name}"
            SET collection_ids = array_remove(collection_ids, :collection_id)
            WHERE :collection_id = ANY(collection_ids)
            """
        )

        with self.vx.Session() as sess:
            sess.execute(query, {"collection_id": collection_id})
            sess.commit()

    def delete_user(self, user_id: str) -> None:
        if self.collection is None:
            raise ValueError("Collection is not initialized.")

        table_name = self.collection.table.name
        query = text(
            f"""
            UPDATE vecs."{table_name}"
            SET user_id = NULL
            WHERE user_id = :user_id
            """
        )

        with self.vx.Session() as sess:
            sess.execute(query, {"user_id": user_id})
            sess.commit()

    def delete_collection(self, collection_id: str) -> None:
        """
        Remove the specified collection ID from all documents in the vector database.

        Args:
            collection_id (str): The ID of the collection to remove from all documents.

        Raises:
            ValueError: If the collection is not initialized.
        """
        if self.collection is None:
            raise ValueError("Collection is not initialized.")

        table_name = self.collection.table.name
        query = text(
            f"""
            UPDATE vecs."{table_name}"
            SET collection_ids = array_remove(collection_ids, :collection_id)
            WHERE :collection_id = ANY(collection_ids)
            """
        )

        with self.vx.Session() as sess:
            result = sess.execute(
                query, {"collection_id": collection_id}
            ).fetchone()
            sess.commit()

        if not result:
            raise ValueError(
                f"Collection {collection_id} not found in any documents."
            )

        affected_rows = result.rowcount
        logger.info(
            f"Removed collection {collection_id} from {affected_rows} documents."
        )

    def get_document_chunks(
        self, document_id: str, offset: int = 0, limit: int = -1
    ) -> dict[str, Any]:
        if not self.collection:
            raise ValueError("Collection is not initialized.")

        limit_clause = f"LIMIT {limit}" if limit != -1 else ""
        table_name = self.collection.table.name
        query = text(
            f"""
            SELECT fragment_id, extraction_id, document_id, user_id, collection_ids, text, metadata, COUNT(*) OVER() AS total
            FROM vecs."{table_name}"
            WHERE document_id = :document_id
            ORDER BY CAST(metadata->>'chunk_order' AS INTEGER)
            {limit_clause} OFFSET :offset
        """
        )

        params = {"document_id": document_id, "offset": offset}
        if limit != -1:
            params["limit"] = limit

        with self.vx.Session() as sess:
            results = sess.execute(query, params).fetchall()

        chunks = []
        total = 0

        if results:
            total = results[0][7]
            chunks = [
                {
                    "fragment_id": result[0],
                    "extraction_id": result[1],
                    "document_id": result[2],
                    "user_id": result[3],
                    "collection_ids": result[4],
                    "text": result[5],
                    "metadata": result[6],
                }
                for result in results
            ]

        return {"results": chunks, "total_entries": total}

    def close(self) -> None:
        if self.vx:
            with self.vx.Session() as sess:
                sess.close()
                if sess.bind:
                    sess.bind.dispose()  # type: ignore

        logger.info("Closed PGVectorDB connection.")



from typing import Any, AsyncGenerator, Dict
from uuid import UUID

from core.base import (
    AsyncPipe,
    AsyncState,
    VectorSearchResult,
    VectorSearchSettings,
)


class RoutingSearchPipe(AsyncPipe):
    def __init__(
        self,
        search_pipes: Dict[str, AsyncPipe],
        default_strategy: str,
        config: AsyncPipe.PipeConfig,
        *args,
        **kwargs,
    ):
        super().__init__(config, *args, **kwargs)
        self.search_pipes = search_pipes
        self.default_strategy = default_strategy

    async def _run_logic(  # type: ignore
        self,
        input: AsyncPipe.Input,
        state: AsyncState,
        run_id: UUID,
        vector_search_settings: VectorSearchSettings,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[VectorSearchResult, None]:
        search_pipe = self.search_pipes.get(
            vector_search_settings.search_strategy
        )
        if not search_pipe:
            raise ValueError(
                f"Search strategy {vector_search_settings.search_strategy} not found"
            )

        async for result in search_pipe._run_logic(  # type: ignore
            input,
            state,
            run_id,
            vector_search_settings=vector_search_settings,
            *args,
            **kwargs,
        ):
            yield result



import logging
import os
from enum import Enum
from pathlib import Path
from typing import Any, Optional

import toml
from pydantic import BaseModel

from ..base.abstractions import GenerationConfig
from ..base.agent.agent import AgentConfig
from ..base.logging.run_logger import LoggingConfig
from ..base.providers.auth import AuthConfig
from ..base.providers.chunking import ChunkingConfig
from ..base.providers.crypto import CryptoConfig
from ..base.providers.database import DatabaseConfig
from ..base.providers.embedding import EmbeddingConfig
from ..base.providers.file import FileConfig
from ..base.providers.kg import KGConfig
from ..base.providers.llm import CompletionConfig
from ..base.providers.parsing import ParsingConfig
from ..base.providers.prompt import PromptConfig

logger = logging.getLogger(__name__)


class R2RConfig:
    current_file_path = os.path.dirname(__file__)
    config_dir_root = os.path.join(current_file_path, "..", "configs")
    default_config_path = os.path.join(
        current_file_path, "..", "..", "r2r.toml"
    )

    CONFIG_OPTIONS: dict[str, Optional[str]] = {}
    for file_ in os.listdir(config_dir_root):
        if file_.endswith(".toml"):
            CONFIG_OPTIONS[file_.removesuffix(".toml")] = os.path.join(
                config_dir_root, file_
            )
    CONFIG_OPTIONS["default"] = None

    REQUIRED_KEYS: dict[str, list] = {
        "crypto": ["provider"],
        "auth": ["provider"],
        "embedding": [
            "provider",
            "base_model",
            "base_dimension",
            "batch_size",
            "add_title_as_prefix",
        ],
        "kg": [
            "provider",
            "batch_size",
            "kg_enrichment_settings",
        ],
        "parsing": ["provider", "excluded_parsers"],
        "chunking": ["provider"],
        "completion": ["provider"],
        "logging": ["provider", "log_table"],
        "prompt": ["provider"],
        "database": ["provider"],
        "agent": ["generation_config"],
        "file": ["provider"],
    }
    auth: AuthConfig
    chunking: ChunkingConfig
    completion: CompletionConfig
    crypto: CryptoConfig
    database: DatabaseConfig
    embedding: EmbeddingConfig
    kg: KGConfig
    logging: LoggingConfig
    parsing: ParsingConfig
    prompt: PromptConfig
    agent: AgentConfig
    file: FileConfig

    def __init__(
        self, config_data: dict[str, Any], base_path: Optional[Path] = None
    ):
        """
        :param config_data: dictionary of configuration parameters
        :param base_path: base path when a relative path is specified for the prompts directory
        """
        # Load the default configuration
        default_config = self.load_default_config()

        # Override the default configuration with the passed configuration
        for key in config_data:
            if key in default_config:
                default_config[key].update(config_data[key])
            else:
                default_config[key] = config_data[key]

        # Validate and set the configuration
        for section, keys in R2RConfig.REQUIRED_KEYS.items():
            # Check the keys when provider is set
            # TODO - Clean up robust null checks
            if "provider" in default_config[section] and (
                default_config[section]["provider"] is not None
                and default_config[section]["provider"] != "None"
                and default_config[section]["provider"] != "null"
            ):
                self._validate_config_section(default_config, section, keys)
                if (
                    section == "prompt"
                    and "file_path" in default_config[section]
                    and not Path(
                        default_config[section]["file_path"]
                    ).is_absolute()
                    and base_path
                ):
                    # Make file_path absolute and relative to the base path
                    default_config[section]["file_path"] = str(
                        base_path / default_config[section]["file_path"]
                    )
            setattr(self, section, default_config[section])

        self.completion = CompletionConfig.create(**self.completion)  # type: ignore
        # override GenerationConfig defaults
        GenerationConfig.set_default(
            **self.completion.generation_config.dict()
        )

        self.auth = AuthConfig.create(**self.auth)  # type: ignore
        self.chunking = ChunkingConfig.create(**self.chunking)  # type: ignore
        self.crypto = CryptoConfig.create(**self.crypto)  # type: ignore
        self.database = DatabaseConfig.create(**self.database)  # type: ignore
        self.embedding = EmbeddingConfig.create(**self.embedding)  # type: ignore
        self.kg = KGConfig.create(**self.kg)  # type: ignore
        self.logging = LoggingConfig.create(**self.logging)  # type: ignore
        if "chunking_config" not in self.parsing:  # type: ignore
            self.parsing["chunking_config"] = self.chunking  # type: ignore
        self.parsing = ParsingConfig.create(**self.parsing)  # type: ignore
        self.prompt = PromptConfig.create(**self.prompt)  # type: ignore
        self.agent = AgentConfig.create(**self.agent)  # type: ignore
        self.file = FileConfig.create(**self.file)  # type: ignore

    def _validate_config_section(
        self, config_data: dict[str, Any], section: str, keys: list
    ):
        if section not in config_data:
            raise ValueError(f"Missing '{section}' section in config")
        if missing_keys := [
            key for key in keys if key not in config_data[section]
        ]:
            raise ValueError(
                f"Missing required keys in '{section}' config: {', '.join(missing_keys)}"
            )

    @classmethod
    def from_toml(cls, config_path: Optional[str] = None) -> "R2RConfig":
        if config_path is None:
            config_path = R2RConfig.default_config_path

        # Load configuration from TOML file
        with open(config_path) as f:
            config_data = toml.load(f)

        return cls(config_data, base_path=Path(config_path).parent)

    def to_toml(self):
        config_data = {
            section: self._serialize_config(getattr(self, section))
            for section in R2RConfig.REQUIRED_KEYS.keys()
        }
        return toml.dumps(config_data)

    def save_to_redis(self, redis_client: Any, key: str):
        redis_client.set(f"R2RConfig:{key}", self.to_toml())

    @classmethod
    def load_from_redis(cls, redis_client: Any, key: str) -> "R2RConfig":
        config_data = redis_client.get(f"R2RConfig:{key}")
        if config_data is None:
            raise ValueError(
                f"Configuration not found in Redis with key '{key}'"
            )
        config_data = toml.loads(config_data)
        return cls(config_data)

    @classmethod
    def load_default_config(cls) -> dict:
        with open(R2RConfig.default_config_path) as f:
            return toml.load(f)

    @staticmethod
    def _serialize_config(config_section: Any) -> dict:
        if isinstance(config_section, dict):
            return {
                R2RConfig._serialize_key(k): R2RConfig._serialize_config(v)
                for k, v in config_section.items()
            }
        elif isinstance(config_section, (list, tuple)):
            return [  # type: ignore
                R2RConfig._serialize_config(item) for item in config_section
            ]
        elif isinstance(config_section, Enum):
            return config_section.value
        elif isinstance(config_section, BaseModel):
            return R2RConfig._serialize_config(
                config_section.model_dump(exclude_none=True)
            )
        else:
            return config_section

    @staticmethod
    def _serialize_key(key: Any) -> str:
        return key.value if isinstance(key, Enum) else str(key)

    @classmethod
    def load(
        cls,
        config_name: Optional[str] = None,
        config_path: Optional[str] = None,
    ) -> "R2RConfig":
        if config_path and config_name:
            raise ValueError("Cannot specify both config_path and config_name")

        if config_path := os.getenv("CONFIG_PATH") or config_path:
            return cls.from_toml(config_path)
        else:
            config_name = os.getenv("CONFIG_NAME") or config_name or "default"
            if config_name not in R2RConfig.CONFIG_OPTIONS:
                raise ValueError(f"Invalid config name: {config_name}")
            return cls.from_toml(R2RConfig.CONFIG_OPTIONS[config_name])



from typing import AsyncGenerator, Optional, Union

from .models import (
    GenerationConfig,
    KGSearchSettings,
    Message,
    RAGResponse,
    SearchResponse,
    VectorSearchSettings,
)


class RetrievalMethods:
    @staticmethod
    async def search(
        client,
        query: str,
        vector_search_settings: Optional[
            Union[dict, VectorSearchSettings]
        ] = None,
        kg_search_settings: Optional[Union[dict, KGSearchSettings]] = None,
    ) -> SearchResponse:
        """
        Conduct a vector and/or KG search.

        Args:
            query (str): The query to search for.
            vector_search_settings (Optional[Union[dict, VectorSearchSettings]]): Vector search settings.
            kg_search_settings (Optional[Union[dict, KGSearchSettings]]): KG search settings.

        Returns:
            SearchResponse: The search response.
        """
        if vector_search_settings and not isinstance(
            vector_search_settings, dict
        ):
            vector_search_settings = vector_search_settings.model_dump()
        if kg_search_settings and not isinstance(kg_search_settings, dict):
            kg_search_settings = kg_search_settings.model_dump()

        data = {
            "query": query,
            "vector_search_settings": vector_search_settings,
            "kg_search_settings": kg_search_settings,
        }
        return await client._make_request("POST", "search", json=data)

    @staticmethod
    async def rag(
        client,
        query: str,
        rag_generation_config: Optional[Union[dict, GenerationConfig]] = None,
        vector_search_settings: Optional[
            Union[dict, VectorSearchSettings]
        ] = None,
        kg_search_settings: Optional[Union[dict, KGSearchSettings]] = None,
        task_prompt_override: Optional[str] = None,
        include_title_if_available: Optional[bool] = False,
    ) -> Union[RAGResponse, AsyncGenerator[RAGResponse, None]]:
        """
        Conducts a Retrieval Augmented Generation (RAG) search with the given query.

        Args:
            query (str): The query to search for.
            rag_generation_config (Optional[Union[dict, GenerationConfig]]): RAG generation configuration.
            vector_search_settings (Optional[Union[dict, VectorSearchSettings]]): Vector search settings.
            kg_search_settings (Optional[Union[dict, KGSearchSettings]]): KG search settings.
            task_prompt_override (Optional[str]): Task prompt override.
            include_title_if_available (Optional[bool]): Include the title if available.

        Returns:
            Union[RAGResponse, AsyncGenerator[RAGResponse, None]]: The RAG response
        """
        if rag_generation_config and not isinstance(
            rag_generation_config, dict
        ):
            rag_generation_config = rag_generation_config.model_dump()
        if vector_search_settings and not isinstance(
            vector_search_settings, dict
        ):
            vector_search_settings = vector_search_settings.model_dump()
        if kg_search_settings and not isinstance(kg_search_settings, dict):
            kg_search_settings = kg_search_settings.model_dump()

        data = {
            "query": query,
            "rag_generation_config": rag_generation_config,
            "vector_search_settings": vector_search_settings,
            "kg_search_settings": kg_search_settings,
            "task_prompt_override": task_prompt_override,
            "include_title_if_available": include_title_if_available,
        }

        if rag_generation_config and rag_generation_config.get(  # type: ignore
            "stream", False
        ):
            return client._make_streaming_request("POST", "rag", json=data)
        else:
            return await client._make_request("POST", "rag", json=data)

    @staticmethod
    async def agent(
        client,
        messages: list[Union[dict, Message]],
        rag_generation_config: Optional[Union[dict, GenerationConfig]],
        vector_search_settings: Optional[
            Union[dict, VectorSearchSettings]
        ] = None,
        kg_search_settings: Optional[Union[dict, KGSearchSettings]] = None,
        task_prompt_override: Optional[str] = None,
        include_title_if_available: Optional[bool] = False,
    ) -> Union[list[Message], AsyncGenerator[Message, None]]:
        """
        Performs a single turn in a conversation with a RAG agent.

        Args:
            messages (List[Union[dict, Message]]): The messages to send to the agent.
            rag_generation_config (Optional[Union[dict, GenerationConfig]]): RAG generation configuration.
            vector_search_settings (Optional[Union[dict, VectorSearchSettings]]): Vector search settings.
            kg_search_settings (Optional[Union[dict, KGSearchSettings]]): KG search settings.
            task_prompt_override (Optional[str]): Task prompt override.
            include_title_if_available (Optional[bool]): Include the title if available.

        Returns:
            Union[List[Message], AsyncGenerator[Message, None]]: The agent response.
        """
        if rag_generation_config and not isinstance(
            rag_generation_config, dict
        ):
            rag_generation_config = rag_generation_config.model_dump()
        if vector_search_settings and not isinstance(
            vector_search_settings, dict
        ):
            vector_search_settings = vector_search_settings.model_dump()
        if kg_search_settings and not isinstance(kg_search_settings, dict):
            kg_search_settings = kg_search_settings.model_dump()

        cast_messages: list[Message] = [
            Message(**msg) if isinstance(msg, dict) else msg
            for msg in messages
        ]

        data = {
            "messages": [msg.model_dump() for msg in cast_messages],
            "rag_generation_config": rag_generation_config,
            "vector_search_settings": vector_search_settings,
            "kg_search_settings": kg_search_settings,
            "task_prompt_override": task_prompt_override,
            "include_title_if_available": include_title_if_available,
        }

        if rag_generation_config and rag_generation_config.get(  # type: ignore
            "stream", False
        ):
            return client._make_streaming_request("POST", "agent", json=data)
        else:
            return await client._make_request("POST", "agent", json=data)



import base64
import logging
import os
from typing import AsyncGenerator

from core.base.abstractions import DataType
from core.base.parsers.base_parser import AsyncParser
from core.parsers.media.openai_helpers import process_frame_with_openai

logger = logging.getLogger(__name__)


class ImageParser(AsyncParser[DataType]):
    """A parser for image data."""

    def __init__(
        self,
        model: str = "gpt-4o-mini",
        max_tokens: int = 2_048,
        api_base: str = "https://api.openai.com/v1/chat/completions",
        max_image_size: int = 1 * 1024 * 1024,  # 4MB limit
    ):
        self.model = model
        self.max_tokens = max_tokens
        self.openai_api_key = os.environ.get("OPENAI_API_KEY")
        self.api_base = api_base
        self.max_image_size = max_image_size

    async def ingest(  # type: ignore
        self, data: DataType, chunk_size: int = 1024
    ) -> AsyncGenerator[str, None]:
        """Ingest image data and yield a description."""

        if isinstance(data, bytes):
            # Encode to base64
            data = base64.b64encode(data).decode("utf-8")

        openai_text = process_frame_with_openai(
            data,  # type: ignore
            self.openai_api_key,  # type: ignore
            self.model,
            self.max_tokens,
            self.api_base,
        )

        # split text into small chunks and yield them
        for i in range(0, len(openai_text), chunk_size):
            text = openai_text[i : i + chunk_size]
            if text and text != "":
                yield text



from datetime import datetime
from typing import Any, Optional
from uuid import UUID

from pydantic import BaseModel

from shared.api.models.base import PaginatedResultsWrapper, ResultsWrapper


class UpdatePromptResponse(BaseModel):
    message: str


class PromptResponse(BaseModel):
    name: str
    template: str
    created_at: datetime
    updated_at: datetime
    input_types: dict[str, str]


class AllPromptsResponse(BaseModel):
    prompts: dict[str, PromptResponse]


class LogEntry(BaseModel):
    key: str
    value: Any
    timestamp: datetime


class LogResponse(BaseModel):
    run_id: UUID
    run_type: str
    entries: list[LogEntry]
    timestamp: Optional[datetime]
    user_id: Optional[UUID]


class ServerStats(BaseModel):
    start_time: datetime
    uptime_seconds: float
    cpu_usage: float
    memory_usage: float


class AnalyticsResponse(BaseModel):
    analytics_data: Optional[dict] = None
    filtered_logs: dict[str, Any]


class AppSettingsResponse(BaseModel):
    config: dict[str, Any]
    prompts: dict[str, Any]


class ScoreCompletionResponse(BaseModel):
    message: str


class UserOverviewResponse(BaseModel):
    user_id: UUID
    num_files: int
    total_size_in_bytes: int
    document_ids: list[UUID]


class UserResponse(BaseModel):
    id: UUID
    email: str
    is_active: bool = True
    is_superuser: bool = False
    created_at: datetime = datetime.now()
    updated_at: datetime = datetime.now()
    is_verified: bool = False
    collection_ids: list[UUID] = []

    # Optional fields (to update or set at creation)
    hashed_password: Optional[str] = None
    verification_code_expiry: Optional[datetime] = None
    name: Optional[str] = None
    bio: Optional[str] = None
    profile_picture: Optional[str] = None


class DocumentOverviewResponse(BaseModel):
    id: UUID
    title: str
    user_id: UUID
    type: str
    created_at: datetime
    updated_at: datetime
    ingestion_status: str
    restructuring_status: str
    version: str
    collection_ids: list[UUID]
    metadata: dict[str, Any]


class DocumentChunkResponse(BaseModel):
    fragment_id: UUID
    extraction_id: UUID
    document_id: UUID
    user_id: UUID
    collection_ids: list[UUID]
    text: str
    metadata: dict[str, Any]


KnowledgeGraphResponse = str


class CollectionResponse(BaseModel):
    collection_id: UUID
    name: str
    description: Optional[str]
    created_at: datetime
    updated_at: datetime


class CollectionOverviewResponse(BaseModel):
    collection_id: UUID
    name: str
    description: Optional[str]
    created_at: datetime
    updated_at: datetime
    user_count: int
    document_count: int


class AddUserResponse(BaseModel):
    result: bool


# Create wrapped versions of each response
WrappedPromptMessageResponse = ResultsWrapper[UpdatePromptResponse]
WrappedGetPromptsResponse = ResultsWrapper[AllPromptsResponse]
WrappedServerStatsResponse = ResultsWrapper[ServerStats]
WrappedLogResponse = ResultsWrapper[list[LogResponse]]
WrappedAnalyticsResponse = ResultsWrapper[AnalyticsResponse]
WrappedAppSettingsResponse = ResultsWrapper[AppSettingsResponse]
WrappedScoreCompletionResponse = ResultsWrapper[ScoreCompletionResponse]
WrappedUserOverviewResponse = PaginatedResultsWrapper[
    list[UserOverviewResponse]
]
WrappedDocumentOverviewResponse = PaginatedResultsWrapper[
    list[DocumentOverviewResponse]
]
WrappedKnowledgeGraphResponse = ResultsWrapper[KnowledgeGraphResponse]
WrappedCollectionResponse = ResultsWrapper[CollectionResponse]
WrappedCollectionListResponse = ResultsWrapper[list[CollectionResponse]]
WrappedCollectionOverviewResponse = ResultsWrapper[
    list[CollectionOverviewResponse]
]
WrappedAddUserResponse = ResultsWrapper[AddUserResponse]
WrappedUsersInCollectionResponse = PaginatedResultsWrapper[list[UserResponse]]
WrappedUserCollectionResponse = PaginatedResultsWrapper[
    list[CollectionOverviewResponse]
]
WrappedDocumentChunkResponse = PaginatedResultsWrapper[
    list[DocumentChunkResponse]
]



# type: ignore
import logging
from typing import Any, AsyncGenerator, Union

from core.base import (
    ChunkingProvider,
    R2RChunkingConfig,
    RecursiveCharacterTextSplitter,
    Strategy,
    TextSplitter,
)
from core.base.abstractions import DocumentExtraction

logger = logging.getLogger(__name__)


class R2RChunkingProvider(ChunkingProvider):
    def __init__(self, config: R2RChunkingConfig):
        super().__init__(config)
        self.text_splitter = self._initialize_text_splitter()
        logger.info(
            f"R2RChunkingProvider initialized with config: {self.config}"
        )

    def _initialize_text_splitter(self) -> TextSplitter:
        logger.info(
            f"Initializing text splitter with method: {self.config.method}"
        )  # Debug log
        if self.config.method == Strategy.RECURSIVE:
            return RecursiveCharacterTextSplitter(
                chunk_size=self.config.chunk_size,
                chunk_overlap=self.config.chunk_overlap,
            )
        elif self.config.method == Strategy.CHARACTER:
            from core.base.utils.splitter.text import CharacterTextSplitter

            separator = CharacterTextSplitter.DEFAULT_SEPARATOR
            if self.config.extra_fields:
                separator = self.config.extra_fields.get(
                    "separator", CharacterTextSplitter.DEFAULT_SEPARATOR
                )
            return CharacterTextSplitter(
                chunk_size=self.config.chunk_size,
                chunk_overlap=self.config.chunk_overlap,
                separator=separator,
                keep_separator=False,
                strip_whitespace=True,
            )
        elif self.config.method == Strategy.BASIC:
            raise NotImplementedError(
                "Basic chunking method not implemented. Please use Recursive."
            )
        elif self.config.method == Strategy.BY_TITLE:
            raise NotImplementedError("By title method not implemented")
        else:
            raise ValueError(f"Unsupported method type: {self.config.method}")

    def validate_config(self) -> bool:
        return self.config.chunk_size > 0 and self.config.chunk_overlap >= 0

    def update_config(self, config_override: R2RChunkingConfig):
        if self.config != config_override:
            self.config = config_override
            self.text_splitter = self._initialize_text_splitter()

    async def chunk(
        self, parsed_document: Union[str, DocumentExtraction]
    ) -> AsyncGenerator[Any, None]:

        if isinstance(parsed_document, DocumentExtraction):
            parsed_document = parsed_document.data

        if isinstance(parsed_document, str):
            chunks = self.text_splitter.create_documents([parsed_document])
        else:
            # Assuming parsed_document is already a list of text chunks
            chunks = parsed_document

        for chunk in chunks:
            yield (
                chunk.page_content if hasattr(chunk, "page_content") else chunk
            )

    async def chunk_with_override(
        self,
        parsed_document: Union[str, DocumentExtraction],
        config_override: R2RChunkingConfig,
    ) -> AsyncGenerator[Any, None]:
        original_config = self.config
        original_splitter = self.text_splitter
        try:
            self.update_config(config_override)
            async for chunk in self.chunk(parsed_document):
                yield chunk
        finally:
            self.config = original_config
            self.text_splitter = original_splitter

    @classmethod
    def with_override(
        cls,
        original_provider: "R2RChunkingProvider",
        config_override: R2RChunkingConfig,
    ) -> "R2RChunkingProvider":
        new_config = R2RChunkingConfig(**original_provider.config.model_dump())
        new_config.update(config_override.model_dump(exclude_unset=True))
        return cls(new_config)



# type: ignore
from typing import AsyncGenerator

from core.base.abstractions import DataType
from core.base.parsers.base_parser import AsyncParser


class TextParser(AsyncParser[DataType]):
    """A parser for raw text data."""

    async def ingest(self, data: DataType) -> AsyncGenerator[DataType, None]:
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        yield data



import os
import shutil
import subprocess

import asyncclick as click

from cli.command_group import cli

REPO_URL = "https://github.com/SciPhi-AI/R2R.git"
TEMPLATES_DIR = "templates"


def get_templates():
    """Fetch list of available templates."""
    temp_dir = "temp_repo"
    try:
        _prepare_temp_directory(temp_dir)
        _clone_and_checkout_templates(temp_dir)
        return _get_template_list()
    except subprocess.CalledProcessError as e:
        raise click.ClickException(
            f"Failed to fetch templates: {e.stderr}"
        ) from e
    except Exception as e:
        raise click.ClickException(
            f"An unexpected error occurred: {str(e)}"
        ) from e
    finally:
        _cleanup(temp_dir)


def _prepare_temp_directory(temp_dir):
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.makedirs(temp_dir, exist_ok=True)


def _clone_and_checkout_templates(temp_dir):
    subprocess.run(
        [
            "git",
            "clone",
            "--depth",
            "1",
            "--filter=blob:none",
            "--sparse",
            REPO_URL,
            temp_dir,
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    os.chdir(temp_dir)
    subprocess.run(
        ["git", "sparse-checkout", "set", TEMPLATES_DIR],
        check=True,
        capture_output=True,
        text=True,
    )
    subprocess.run(
        ["git", "checkout"],
        check=True,
        capture_output=True,
        text=True,
    )


def _get_template_list():
    if not os.path.exists(TEMPLATES_DIR):
        raise click.ClickException(
            f"Templates directory '{TEMPLATES_DIR}' not found in the repository."
        )

    if templates := [
        d
        for d in os.listdir(TEMPLATES_DIR)
        if os.path.isdir(os.path.join(TEMPLATES_DIR, d))
    ]:
        return templates
    else:
        raise click.ClickException("No templates found in the repository.")


def _cleanup(temp_dir):
    os.chdir("..")
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)


def clone_template(template_name, location):
    templates = get_templates()
    if template_name not in templates:
        raise ValueError(
            f"Template '{template_name}' not found. Available templates: {', '.join(templates)}"
        )

    if not location:
        location = template_name

    if os.path.exists(location) and os.listdir(location):
        raise ValueError(
            f"Directory '{location}' already exists and is not empty."
        )

    # Clone the repository
    subprocess.run(
        [
            "git",
            "clone",
            "--depth",
            "1",  # Shallow clone
            "--filter=blob:none",  # Don't download file contents initially
            "--sparse",  # Enable sparse checkout
            REPO_URL,
            location,
        ],
        check=True,
    )

    os.chdir(location)
    subprocess.run(
        [
            "git",
            "sparse-checkout",
            "set",
            f"{TEMPLATES_DIR}/{template_name}",
        ],
        check=True,
    )
    subprocess.run(["git", "checkout"], check=True)

    template_dir = os.path.join(TEMPLATES_DIR, template_name)
    if not os.path.exists(template_dir):
        raise ValueError(
            f"Template directory '{template_name}' not found in the cloned repository."
        )

    for item in os.listdir(template_dir):
        shutil.move(os.path.join(template_dir, item), item)

    shutil.rmtree(TEMPLATES_DIR)
    shutil.rmtree(".git")


@cli.command()
@click.argument("template_name", required=True)
@click.argument("location", required=False)
def clone(template_name, location):
    """Clones a template repository."""
    try:
        clone_template(template_name, location)
        click.echo(
            f"Successfully cloned template '{template_name}' to {location or template_name}"
        )
    except ValueError as e:
        raise click.ClickException(str(e)) from e
    except subprocess.CalledProcessError as e:
        raise click.ClickException(f"Git operation failed: {e}") from e
    except Exception as e:
        raise click.ClickException(
            f"An unexpected error occurred: {str(e)}"
        ) from e


@cli.command()
def list_templates():
    """Lists all available templates."""
    try:
        templates = get_templates()
        click.echo("Available templates:")
        for template in templates:
            click.echo(f"- {template}")
    except Exception as e:
        click.echo(f"Error: {str(e)}", err=True)
        raise click.ClickException(str(e)) from e



from .abstractions.generator_pipe import GeneratorPipe
from .abstractions.search_pipe import SearchPipe
from .ingestion.chunking_pipe import ChunkingPipe
from .ingestion.embedding_pipe import EmbeddingPipe
from .ingestion.parsing_pipe import ParsingPipe
from .ingestion.vector_storage_pipe import VectorStoragePipe
from .kg.clustering import KGClusteringPipe
from .kg.community_summary import KGCommunitySummaryPipe
from .kg.extraction import KGTriplesExtractionPipe
from .kg.node_extraction import KGNodeDescriptionPipe, KGNodeExtractionPipe
from .kg.storage import KGStoragePipe
from .retrieval.kg_search_pipe import KGSearchSearchPipe
from .retrieval.multi_search import MultiSearchPipe
from .retrieval.query_transform_pipe import QueryTransformPipe
from .retrieval.routing_search_pipe import RoutingSearchPipe
from .retrieval.search_rag_pipe import SearchRAGPipe
from .retrieval.streaming_rag_pipe import StreamingSearchRAGPipe
from .retrieval.vector_search_pipe import VectorSearchPipe

__all__ = [
    "SearchPipe",
    "GeneratorPipe",
    "EmbeddingPipe",
    "KGTriplesExtractionPipe",
    "KGNodeExtractionPipe",
    "KGSearchSearchPipe",
    "KGNodeDescriptionPipe",
    "ParsingPipe",
    "ChunkingPipe",
    "QueryTransformPipe",
    "SearchRAGPipe",
    "StreamingSearchRAGPipe",
    "VectorSearchPipe",
    "VectorStoragePipe",
    "KGStoragePipe",
    "KGClusteringPipe",
    "MultiSearchPipe",
    "KGCommunitySummaryPipe",
    "RoutingSearchPipe",
]



import asyncio
import os
import uuid
from concurrent.futures import ThreadPoolExecutor

import aiofiles
from datasets import load_dataset

from r2r import R2RClient


def generate_id_from_label(label: str) -> uuid.UUID:
    return uuid.uuid5(uuid.NAMESPACE_DNS, label)


async def upload_batch(client, file_paths, executor):
    print(f"Uploading batch of {len(file_paths)} files")
    loop = asyncio.get_event_loop()
    try:
        # Assuming ingest_files is a blocking I/O operation
        response = await loop.run_in_executor(
            executor, client.ingest_files, file_paths
        )
        print(response)
    except Exception as e:
        print(f"Error uploading batch: {e}")
    finally:
        # Remove files asynchronously
        await asyncio.gather(*(remove_file(fname) for fname in file_paths))


async def remove_file(file_path):
    try:
        os.remove(file_path)
    except Exception as e:
        print(f"Error removing file {file_path}: {e}")


async def process_dataset(client, dataset, batch_size, executor, semaphore):
    current_batch = []
    count = 0

    for example in dataset:
        count += 1
        fname = f"example_{generate_id_from_label(example['completion'])}.txt"
        print(f"Streaming {fname} w/ completion {count} ...")

        # Asynchronously write to file
        async with aiofiles.open(fname, "w") as f:
            await f.write(example["completion"])

        current_batch.append(fname)

        if len(current_batch) >= batch_size:
            # Acquire semaphore before launching a new upload task
            await semaphore.acquire()
            asyncio.create_task(
                upload_and_release(
                    semaphore,
                    upload_batch,
                    client,
                    current_batch.copy(),
                    executor,
                )
            )
            current_batch.clear()

    # Upload any remaining files
    if current_batch:
        await semaphore.acquire()
        asyncio.create_task(
            upload_and_release(
                semaphore, upload_batch, client, current_batch.copy(), executor
            )
        )


async def upload_and_release(semaphore, upload_func, client, batch, executor):
    try:
        await upload_func(client, batch, executor)
    finally:
        semaphore.release()


async def main():
    r2r_url = os.getenv("R2R_API_URL", "http://localhost:7272")
    print(f"Using R2R API at: {r2r_url}")
    client = R2RClient(r2r_url)

    # Load the dataset
    dataset = load_dataset(
        "SciPhi/textbooks-are-all-you-need-lite", streaming=True
    )["train"]

    batch_size = 64
    max_concurrent_uploads = 10  # Adjust based on your system/API limits
    semaphore = asyncio.Semaphore(max_concurrent_uploads)

    # Use a ThreadPoolExecutor for blocking I/O operations
    with ThreadPoolExecutor(max_workers=max_concurrent_uploads) as executor:
        await process_dataset(client, dataset, batch_size, executor, semaphore)

        # Wait for all upload tasks to complete
        # This ensures that all tasks have been launched
        # and then waits for them to finish
        await asyncio.sleep(0.1)  # Let asyncio schedule the tasks

        # Optionally, gather all pending tasks
        pending = [
            task
            for task in asyncio.all_tasks()
            if task is not asyncio.current_task()
        ]
        if pending:
            await asyncio.gather(*pending)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import json
import logging
from typing import TYPE_CHECKING, Any, AsyncGenerator, Iterable
from uuid import NAMESPACE_DNS, UUID, uuid4, uuid5

from ..abstractions.graph import EntityType, RelationshipType
from ..abstractions.search import AggregateSearchResult

logger = logging.getLogger(__name__)


def format_search_results_for_llm(
    results: AggregateSearchResult,
) -> str:
    formatted_results = ""
    if results.vector_search_results:
        formatted_results += "Vector Search Results:\n"
        for i, result in enumerate(results.vector_search_results):
            text = result.text
            formatted_results += f"Source [{i+1}]:\n{text}\n"

    if results.kg_search_results:
        formatted_results += "KG Local Results:\n"
        for j, kg_result in enumerate(results.kg_search_results):
            formatted_results += (
                f"Source [{j+i+1}]: Name - {kg_result.content.name}\n"
            )
            formatted_results += (
                f"Description - {kg_result.content.description}\n"
            )
            findings = kg_result.metadata.get("findings", None)
            if findings:
                formatted_results += f"Supporting Findings: {findings}\n"

    return formatted_results


def format_search_results_for_stream(
    result: AggregateSearchResult,
) -> str:
    VECTOR_SEARCH_STREAM_MARKER = (
        "search"  # TODO - change this to vector_search in next major release
    )
    KG_SEARCH_STREAM_MARKER = "kg_search"

    context = ""
    if result.vector_search_results:
        context += f"<{VECTOR_SEARCH_STREAM_MARKER}>"
        vector_results_list = [
            result.as_dict() for result in result.vector_search_results
        ]
        context += json.dumps(vector_results_list, default=str)
        context += f"</{VECTOR_SEARCH_STREAM_MARKER}>"

    if result.kg_search_results:
        context += f"<{KG_SEARCH_STREAM_MARKER}>"
        kg_results_list = [
            result.dict() for result in result.kg_search_results
        ]
        context += json.dumps(kg_results_list, default=str)
        context += f"</{KG_SEARCH_STREAM_MARKER}>"

    return context


if TYPE_CHECKING:
    from ..pipeline.base_pipeline import AsyncPipeline


def generate_run_id() -> UUID:
    return uuid5(NAMESPACE_DNS, str(uuid4()))


def generate_id_from_label(label: str) -> UUID:
    return uuid5(NAMESPACE_DNS, label)


def generate_user_document_id(filename: str, user_id: UUID) -> UUID:
    """
    Generates a unique document id from a given filename and user id
    """
    return generate_id_from_label(f'{filename.split("/")[-1]}-{str(user_id)}')


async def to_async_generator(
    iterable: Iterable[Any],
) -> AsyncGenerator[Any, None]:
    for item in iterable:
        yield item


def run_pipeline(pipeline: "AsyncPipeline", input: Any, *args, **kwargs):
    if not isinstance(input, AsyncGenerator):
        if not isinstance(input, list):
            input = to_async_generator([input])
        else:
            input = to_async_generator(input)

    async def _run_pipeline(input, *args, **kwargs):
        return await pipeline.run(input, *args, **kwargs)

    return asyncio.run(_run_pipeline(input, *args, **kwargs))


def increment_version(version: str) -> str:
    prefix = version[:-1]
    suffix = int(version[-1])
    return f"{prefix}{suffix + 1}"


def decrement_version(version: str) -> str:
    prefix = version[:-1]
    suffix = int(version[-1])
    return f"{prefix}{max(0, suffix - 1)}"


def format_entity_types(entity_types: list[EntityType]) -> str:
    lines = [entity.name for entity in entity_types]
    return "\n".join(lines)


def format_relations(predicates: list[RelationshipType]) -> str:
    lines = [predicate.name for predicate in predicates]
    return "\n".join(lines)



from typing import Generic, TypeVar

from pydantic import BaseModel

T = TypeVar("T")


class ResultsWrapper(BaseModel, Generic[T]):
    results: T


class PaginatedResultsWrapper(BaseModel, Generic[T]):
    results: T
    total_entries: int



import asyncio
import json
import logging
import uuid

from hatchet_sdk import ConcurrencyLimitStrategy, Context

from core import GenerationConfig, IngestionStatus, KGCreationSettings
from core.base import R2RDocumentProcessingError
from core.base.abstractions import RestructureStatus

from ..services import RestructureService
from .base import r2r_hatchet

logger = logging.getLogger(__name__)


@r2r_hatchet.workflow(name="kg-extract-and-store", timeout="60m")
class KgExtractAndStoreWorkflow:
    def __init__(self, restructure_service: RestructureService):
        self.restructure_service = restructure_service

    @r2r_hatchet.step(retries=3, timeout="60m")
    async def kg_extract_and_store(self, context: Context) -> dict:
        input_data = context.workflow_input()["request"]
        document_id = uuid.UUID(input_data["document_id"])
        fragment_merge_count = input_data["fragment_merge_count"]
        max_knowledge_triples = input_data["max_knowledge_triples"]
        entity_types = input_data["entity_types"]
        relation_types = input_data["relation_types"]

        document_overview = await self.restructure_service.providers.database.relational.get_documents_overview(
            filter_document_ids=[document_id]
        )
        document_overview = document_overview["results"][0]

        try:

            # Set restructure status to 'processing'
            document_overview.restructuring_status = (
                RestructureStatus.PROCESSING
            )

            await self.restructure_service.providers.database.relational.upsert_documents_overview(
                document_overview
            )

            errors = await self.restructure_service.kg_extract_and_store(
                document_id=document_id,
                generation_config=GenerationConfig(
                    **input_data["generation_config"]
                ),
                fragment_merge_count=fragment_merge_count,
                max_knowledge_triples=max_knowledge_triples,
                entity_types=entity_types,
                relation_types=relation_types,
            )
            # Set restructure status to 'success' if completed successfully
            if len(errors) == 0:
                document_overview.restructuring_status = (
                    RestructureStatus.SUCCESS
                )
                await self.restructure_service.providers.database.relational.upsert_documents_overview(
                    document_overview
                )
            else:

                document_overview.restructuring_status = (
                    RestructureStatus.FAILURE
                )
                await self.restructure_service.providers.database.relational.upsert_documents_overview(
                    document_overview
                )
                raise R2RDocumentProcessingError(
                    error_message=f"Error in kg_extract_and_store, list of errors: {errors}",
                    document_id=document_id,
                )

        except Exception as e:
            # Set restructure status to 'failure' if an error occurred
            document_overview.restructuring_status = RestructureStatus.FAILURE
            await self.restructure_service.providers.database.relational.upsert_documents_overview(
                document_overview
            )
            raise R2RDocumentProcessingError(
                error_message=e,
                document_id=document_id,
            )

        return {"result": None}


@r2r_hatchet.workflow(name="create-graph", timeout="60m")
class CreateGraphWorkflow:
    def __init__(self, restructure_service: RestructureService):
        self.restructure_service = restructure_service

    @r2r_hatchet.step(retries=1)
    async def kg_extraction_ingress(self, context: Context) -> dict:
        input_data = context.workflow_input()["request"]

        kg_creation_settings = KGCreationSettings(
            **json.loads(input_data["kg_creation_settings"])
        )

        documents_overview = (
            await self.restructure_service.providers.database.relational.get_documents_overview()
        )
        documents_overview = documents_overview["results"]

        document_ids = [
            doc.id
            for doc in documents_overview
            if doc.restructuring_status != IngestionStatus.SUCCESS
        ]

        document_ids = [str(doc_id) for doc_id in document_ids]

        documents_overviews = await self.restructure_service.providers.database.relational.get_documents_overview(
            filter_document_ids=document_ids
        )
        documents_overviews = documents_overviews["results"]

        # Only run if restructuring_status is pending or failure
        filtered_document_ids = []
        for document_overview in documents_overviews:
            restructuring_status = document_overview.restructuring_status
            if restructuring_status in [
                RestructureStatus.PENDING,
                RestructureStatus.FAILURE,
                RestructureStatus.ENRICHMENT_FAILURE,
            ]:
                filtered_document_ids.append(document_overview.id)
            elif restructuring_status == RestructureStatus.SUCCESS:
                logger.warning(
                    f"Graph already created for document ID: {document_overview.id}"
                )
            elif restructuring_status == RestructureStatus.PROCESSING:
                logger.warning(
                    f"Graph creation is already in progress for document ID: {document_overview.id}"
                )
            elif restructuring_status == RestructureStatus.ENRICHED:
                logger.warning(
                    f"Graph is already enriched for document ID: {document_overview.id}"
                )
            else:
                logger.warning(
                    f"Unknown restructuring status for document ID: {document_overview.id}"
                )

        results = []
        for document_id in filtered_document_ids:
            logger.info(
                f"Running Graph Creation Workflow for document ID: {document_id}"
            )
            results.append(
                (
                    context.aio.spawn_workflow(
                        "kg-extract-and-store",
                        {
                            "request": {
                                "document_id": str(document_id),
                                "fragment_merge_count": kg_creation_settings.fragment_merge_count,
                                "max_knowledge_triples": kg_creation_settings.max_knowledge_triples,
                                "generation_config": kg_creation_settings.generation_config.to_dict(),
                                "entity_types": kg_creation_settings.entity_types,
                                "relation_types": kg_creation_settings.relation_types,
                            }
                        },
                        key=f"kg-extract-and-store_{document_id}",
                    )
                )
            )

        if not filtered_document_ids:
            logger.info(
                "No documents to process, either all graphs were created or in progress, or no documents were provided. Skipping graph creation."
            )
            return {"result": "success"}

        logger.info(f"Ran {len(results)} workflows for graph creation")
        results = await asyncio.gather(*results)
        return {"result": "success"}


@r2r_hatchet.workflow(name="enrich-graph", timeout="60m")
class EnrichGraphWorkflow:
    def __init__(self, restructure_service: RestructureService):
        self.restructure_service = restructure_service

    @r2r_hatchet.step(retries=3, timeout="60m")
    async def kg_node_creation(self, context: Context) -> dict:
        input_data = context.workflow_input()["request"]
        max_description_input_length = input_data[
            "max_description_input_length"
        ]
        await self.restructure_service.kg_node_creation(
            max_description_input_length=max_description_input_length
        )
        return {"result": None}

    @r2r_hatchet.step(retries=3, parents=["kg_node_creation"], timeout="60m")
    async def kg_clustering(self, context: Context) -> dict:
        input_data = context.workflow_input()["request"]
        skip_clustering = input_data["skip_clustering"]
        force_enrichment = input_data["force_enrichment"]
        leiden_params = input_data["leiden_params"]
        max_summary_input_length = input_data["max_summary_input_length"]
        generation_config = GenerationConfig(**input_data["generation_config"])

        # todo: check if documets are already being clustered
        # check if any documents are still being restructured, need to explicitly set the force_clustering flag to true to run clustering if documents are still being restructured

        documents_overview = (
            await self.restructure_service.providers.database.relational.get_documents_overview()
        )
        documents_overview = documents_overview["results"]

        if not force_enrichment:
            if any(
                document_overview.restructuring_status
                == RestructureStatus.PROCESSING
                for document_overview in documents_overview
            ):
                logger.error(
                    "Graph creation is still in progress for some documents, skipping enrichment, please set force_enrichment to true if you want to run enrichment anyway"
                )
                return {"result": None}

            if any(
                document_overview.restructuring_status
                == RestructureStatus.ENRICHING
                for document_overview in documents_overview
            ):
                logger.error(
                    "Graph enrichment is still in progress for some documents, skipping enrichment, please set force_enrichment to true if you want to run enrichment anyway"
                )
                return {"result": None}

        for document_overview in documents_overview:
            if document_overview.restructuring_status in [
                RestructureStatus.SUCCESS,
                RestructureStatus.ENRICHMENT_FAILURE,
            ]:
                document_overview.restructuring_status = (
                    RestructureStatus.ENRICHING
                )

        await self.restructure_service.providers.database.relational.upsert_documents_overview(
            documents_overview
        )

        try:
            if not skip_clustering:
                results = await self.restructure_service.kg_clustering(
                    leiden_params, generation_config
                )

                result = results[0]

                # Run community summary workflows
                workflows = []
                for level, community_id in result["intermediate_communities"]:
                    logger.info(
                        f"Running KG Community Summary Workflow for community ID: {community_id} at level {level}"
                    )
                    workflows.append(
                        context.aio.spawn_workflow(
                            "kg-community-summary",
                            {
                                "request": {
                                    "community_id": str(community_id),
                                    "level": level,
                                    "generation_config": generation_config.to_dict(),
                                    "max_summary_input_length": max_summary_input_length,
                                }
                            },
                            key=f"kg-community-summary_{community_id}_{level}",
                        )
                    )

                results = await asyncio.gather(*workflows)
            else:
                logger.info(
                    "Skipping Leiden clustering as skip_clustering is True, also skipping community summary workflows"
                )
                return {"result": None}

        except Exception as e:
            logger.error(f"Error in kg_clustering: {str(e)}", exc_info=True)
            documents_overview = (
                await self.restructure_service.providers.database.relational.get_documents_overview()
            )
            documents_overview = documents_overview["results"]
            for document_overview in documents_overview:
                if (
                    document_overview.restructuring_status
                    == RestructureStatus.ENRICHING
                ):
                    document_overview.restructuring_status = (
                        RestructureStatus.ENRICHMENT_FAILURE
                    )
                    await self.restructure_service.providers.database.relational.upsert_documents_overview(
                        document_overview
                    )
                    logger.error(
                        f"Error in kg_clustering for document {document_overview.id}: {str(e)}"
                    )
            raise e

        finally:

            documents_overview = (
                await self.restructure_service.providers.database.relational.get_documents_overview()
            )
            documents_overview = documents_overview["results"]
            for document_overview in documents_overview:
                if (
                    document_overview.restructuring_status
                    == RestructureStatus.ENRICHING
                ):
                    document_overview.restructuring_status = (
                        RestructureStatus.ENRICHED
                    )

            await self.restructure_service.providers.database.relational.upsert_documents_overview(
                documents_overview
            )
            return {"result": None}


@r2r_hatchet.workflow(name="kg-community-summary", timeout="60m")
class KGCommunitySummaryWorkflow:
    def __init__(self, restructure_service: RestructureService):
        self.restructure_service = restructure_service

    @r2r_hatchet.step(retries=1, timeout="60m")
    async def kg_community_summary(self, context: Context) -> dict:
        input_data = context.workflow_input()["request"]
        community_id = input_data["community_id"]
        level = input_data["level"]
        generation_config = GenerationConfig(**input_data["generation_config"])
        max_summary_input_length = input_data["max_summary_input_length"]
        await self.restructure_service.kg_community_summary(
            community_id=community_id,
            level=level,
            max_summary_input_length=max_summary_input_length,
            generation_config=generation_config,
        )
        return {"result": None}



import asyncio
import logging
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from pathlib import Path

import toml

from core.telemetry.events import ErrorEvent, FeatureUsageEvent
from core.telemetry.posthog import telemetry_client

logger = logging.getLogger(__name__)


class ProductTelemetryClient:
    USER_ID_PATH = str(Path.home() / ".cache" / "r2r" / "telemetry_user_id")
    UNKNOWN_USER_ID = "UNKNOWN"
    _curr_user_id = None
    _version = None

    @property
    def version(self) -> str:
        if self._version is None:
            try:
                pyproject_path = (
                    Path(__file__).parent.parent.parent / "pyproject.toml"
                )
                pyproject_data = toml.load(pyproject_path)
                self._version = pyproject_data["tool"]["poetry"]["version"]
            except Exception as e:
                logger.error(
                    f"Error reading version from pyproject.toml: {str(e)}"
                )
                self._version = "UNKNOWN"
        return self._version

    @property
    def user_id(self) -> str:
        if self._curr_user_id:
            return self._curr_user_id

        try:
            if not os.path.exists(self.USER_ID_PATH):
                os.makedirs(os.path.dirname(self.USER_ID_PATH), exist_ok=True)
                with open(self.USER_ID_PATH, "w") as f:
                    new_user_id = str(uuid.uuid4())
                    f.write(new_user_id)
                self._curr_user_id = new_user_id
            else:
                with open(self.USER_ID_PATH, "r") as f:
                    self._curr_user_id = f.read().strip()
        except Exception:
            self._curr_user_id = self.UNKNOWN_USER_ID
        return self._curr_user_id


product_telemetry_client = ProductTelemetryClient()


def get_project_metadata():
    import platform

    return {
        "os": platform.system(),
        "python_version": platform.python_version(),
        "version": product_telemetry_client.version,
    }


# Create a thread pool with a fixed number of workers
telemetry_thread_pool = ThreadPoolExecutor(max_workers=2)


def telemetry_event(event_name):
    def decorator(func):
        def log_telemetry(event_type, user_id, metadata, error_message=None):
            try:
                if event_type == "feature":
                    telemetry_client.capture(
                        FeatureUsageEvent(
                            user_id=user_id,
                            properties=metadata,
                            feature=event_name,
                        )
                    )
                elif event_type == "error":
                    telemetry_client.capture(
                        ErrorEvent(
                            user_id=user_id,
                            properties=metadata,
                            endpoint=event_name,
                            error_message=error_message,
                        )
                    )
            except Exception as e:
                logger.error(f"Error in telemetry event logging: {str(e)}")

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            metadata = get_project_metadata()
            user_id = product_telemetry_client.user_id

            try:
                result = await func(*args, **kwargs)
                telemetry_thread_pool.submit(
                    log_telemetry, "feature", user_id, metadata
                )
                return result
            except Exception as e:
                telemetry_thread_pool.submit(
                    log_telemetry, "error", user_id, metadata, str(e)
                )
                raise

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            loop = asyncio.get_event_loop()
            if loop.is_running():
                future = asyncio.run_coroutine_threadsafe(
                    async_wrapper(*args, **kwargs), loop
                )
                return future.result()
            else:
                return loop.run_until_complete(async_wrapper(*args, **kwargs))

        return (
            async_wrapper
            if asyncio.iscoroutinefunction(func)
            else sync_wrapper
        )

    return decorator



from .auth import R2RAuthProvider
from .chunking import (  # type: ignore
    R2RChunkingProvider,
    UnstructuredChunkingProvider,
)
from .crypto import BCryptConfig, BCryptProvider
from .database import PostgresDBProvider
from .embeddings import (
    LiteLLMEmbeddingProvider,
    OllamaEmbeddingProvider,
    OpenAIEmbeddingProvider,
)
from .file import PostgresFileProvider
from .kg import Neo4jKGProvider
from .llm import LiteCompletionProvider, OpenAICompletionProvider
from .orchestration import HatchetOrchestrationProvider
from .parsing import R2RParsingProvider, UnstructuredParsingProvider
from .prompts import R2RPromptProvider

__all__ = [
    # Auth
    "R2RAuthProvider",
    # Chunking
    "R2RChunkingProvider",  # type: ignore
    "UnstructuredChunkingProvider",  # type: ignore
    # Crypto
    "BCryptProvider",
    "BCryptConfig",
    # Database
    "PostgresDBProvider",
    # Embeddings
    "LiteLLMEmbeddingProvider",
    "OllamaEmbeddingProvider",
    "OpenAIEmbeddingProvider",
    # File
    "PostgresFileProvider",
    # KG
    "Neo4jKGProvider",
    # Orchestration
    "HatchetOrchestrationProvider",
    # LLM
    "OpenAICompletionProvider",
    "LiteCompletionProvider",
    # Parsing
    "R2RParsingProvider",
    "UnstructuredParsingProvider",
    # Prompts
    "R2RPromptProvider",
]



import json
import logging
from collections import defaultdict
from typing import Any, BinaryIO, Dict, Optional, Tuple
from uuid import UUID

import toml

from core.base import (
    AnalysisTypes,
    LogFilterCriteria,
    LogProcessor,
    Prompt,
    R2RException,
    RunLoggingSingleton,
    RunManager,
    RunType,
)
from core.telemetry.telemetry_decorator import telemetry_event

from ..abstractions import R2RAgents, R2RPipelines, R2RPipes, R2RProviders
from ..config import R2RConfig
from .base import Service

logger = logging.getLogger(__name__)


class ManagementService(Service):
    def __init__(
        self,
        config: R2RConfig,
        providers: R2RProviders,
        pipes: R2RPipes,
        pipelines: R2RPipelines,
        agents: R2RAgents,
        run_manager: RunManager,
        logging_connection: RunLoggingSingleton,
    ):
        super().__init__(
            config,
            providers,
            pipes,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

    @telemetry_event("Logs")
    async def logs(
        self,
        offset: int = 0,
        limit: int = 100,
        run_type_filter: Optional[RunType] = None,
    ):
        if self.logging_connection is None:
            raise R2RException(
                status_code=404, message="Logging provider not found."
            )

        run_info = await self.logging_connection.get_info_logs(
            offset=offset,
            limit=limit,
            run_type_filter=run_type_filter,
        )
        run_ids = [run.run_id for run in run_info]
        if not run_ids:
            return []
        logs = await self.logging_connection.get_logs(run_ids)

        aggregated_logs = []

        for run in run_info:
            run_logs = [log for log in logs if log["run_id"] == run.run_id]
            entries = [
                {
                    "key": log["key"],
                    "value": log["value"],
                    "timestamp": log["timestamp"],
                }
                for log in run_logs
            ][
                ::-1
            ]  # Reverse order so that earliest logged values appear first.

            log_entry = {
                "run_id": str(run.run_id),
                "run_type": run.run_type,
                "entries": entries,
            }

            if run.timestamp:
                log_entry["timestamp"] = run.timestamp.isoformat()

            if hasattr(run, "user_id") and run.user_id is not None:
                log_entry["user_id"] = str(run.user_id)

            aggregated_logs.append(log_entry)

        return aggregated_logs

    @telemetry_event("Analytics")
    async def analytics(
        self,
        filter_criteria: LogFilterCriteria,
        analysis_types: AnalysisTypes,
        *args,
        **kwargs,
    ):
        run_info = await self.logging_connection.get_info_logs(limit=100)
        run_ids = [info.run_id for info in run_info]

        if not run_ids:
            return {
                "analytics_data": "No logs found.",
                "filtered_logs": {},
            }
        logs = await self.logging_connection.get_logs(run_ids=run_ids)

        filters = {}
        if filter_criteria.filters:
            for key, value in filter_criteria.filters.items():
                filters[key] = lambda log, value=value: (
                    any(
                        entry.get("key") == value
                        for entry in log.get("entries", [])
                    )
                    if "entries" in log
                    else log.get("key") == value
                )

        log_processor = LogProcessor(filters)  # type: ignore
        for log in logs:
            if "entries" in log and isinstance(log["entries"], list):
                log_processor.process_log(log)
            elif "key" in log:
                log_processor.process_log(log)
            else:
                logger.warning(
                    f"Skipping log due to missing or malformed 'entries': {log}"
                )

        filtered_logs = dict(log_processor.populations.items())

        analytics_data = {}
        if analysis_types and analysis_types.analysis_types:
            for (
                filter_key,
                analysis_config,
            ) in analysis_types.analysis_types.items():
                if filter_key in filtered_logs:
                    analysis_type = analysis_config[0]
                    if analysis_type == "bar_chart":
                        extract_key = analysis_config[1]
                        analytics_data[filter_key] = (
                            AnalysisTypes.generate_bar_chart_data(
                                filtered_logs[filter_key], extract_key
                            )
                        )
                    elif analysis_type == "basic_statistics":
                        extract_key = analysis_config[1]
                        analytics_data[filter_key] = (
                            AnalysisTypes.calculate_basic_statistics(
                                filtered_logs[filter_key], extract_key
                            )
                        )
                    elif analysis_type == "percentile":
                        extract_key = analysis_config[1]
                        percentile = int(analysis_config[2])
                        analytics_data[filter_key] = (
                            AnalysisTypes.calculate_percentile(
                                filtered_logs[filter_key],
                                extract_key,
                                percentile,
                            )
                        )
                    else:
                        logger.warning(
                            f"Unknown analysis type for filter key '{filter_key}': {analysis_type}"
                        )

        return {
            "analytics_data": analytics_data or None,
            "filtered_logs": filtered_logs,
        }

    @telemetry_event("AppSettings")
    async def app_settings(self, *args: Any, **kwargs: Any):
        prompts = self.providers.prompt.get_all_prompts()
        config_toml = self.config.to_toml()
        config_dict = toml.loads(config_toml)
        return {
            "config": config_dict,
            "prompts": prompts,
        }

    @telemetry_event("ScoreCompletion")
    async def score_completion(
        self,
        message_id: UUID,
        score: float = 0.0,
        run_type_filter: Optional[RunType] = None,
        max_runs: int = 100,
        *args: Any,
        **kwargs: Any,
    ):
        try:
            if self.logging_connection is None:
                raise R2RException(
                    status_code=404, message="Logging provider not found."
                )

            run_info = await self.logging_connection.get_info_logs(
                limit=max_runs,
                run_type_filter=run_type_filter,
            )
            run_ids = [run.run_id for run in run_info]

            logs = await self.logging_connection.get_logs(run_ids)

            for log in logs:
                if log["key"] != "completion_record":
                    continue
                completion_record = log["value"]
                try:
                    completion_dict = json.loads(completion_record)
                except json.JSONDecodeError as e:
                    logger.error(f"Error processing completion record: {e}")
                    continue

                if completion_dict.get("message_id") == str(message_id):
                    bounded_score = round(min(max(score, -1.00), 1.00), 2)
                    updated = await RunLoggingSingleton.score_completion(
                        log["run_id"], message_id, bounded_score
                    )
                    if not updated:
                        logger.error(
                            f"Error updating completion record for message_id: {message_id}"
                        )

        except Exception as e:
            logger.error(f"An error occurred in ascore_completion: {e}")

        return {"message": "Completion scored successfully"}

    @telemetry_event("UsersOverview")
    async def users_overview(
        self,
        user_ids: Optional[list[UUID]] = None,
        offset: int = 0,
        limit: int = 100,
        *args,
        **kwargs,
    ):
        return await self.providers.database.relational.get_users_overview(
            [str(ele) for ele in user_ids] if user_ids else None,
            offset=offset,
            limit=limit,
        )

    @telemetry_event("Delete")
    async def delete(
        self,
        filters: dict[str, Any],
        *args,
        **kwargs,
    ):
        """
        Takes a list of filters like
        "{key: {operator: value}, key: {operator: value}, ...}"
        and deletes entries matching the given filters from both vector and relational databases.

        NOTE: This method is not atomic and may result in orphaned entries in the documents overview table.
        NOTE: This method assumes that filters delete entire contents of any touched documents.
        """
        logger.info(f"Deleting entries with filters: {filters}")

        try:
            vector_delete_results = self.providers.database.vector.delete(
                filters
            )
        except Exception as e:
            logger.error(f"Error deleting from vector database: {e}")
            vector_delete_results = {}

        document_ids_to_purge: set[UUID] = set()
        if vector_delete_results:
            document_ids_to_purge.update(
                doc_id
                for doc_id in (
                    result.get("document_id")
                    for result in vector_delete_results.values()
                )
                if doc_id
            )

        relational_filters = {}
        if "document_id" in filters:
            relational_filters["filter_document_ids"] = [
                UUID(filters["document_id"]["$eq"])
            ]
        if "user_id" in filters:
            relational_filters["filter_user_ids"] = [
                UUID(filters["user_id"]["$eq"])
            ]
        if "collection_ids" in filters:
            relational_filters["filter_collection_ids"] = [
                UUID(collection_id)
                for collection_id in filters["collection_ids"]["$in"]
            ]

        try:
            documents_overview = await self.providers.database.relational.get_documents_overview(
                **relational_filters
            )
        except Exception as e:
            logger.error(
                f"Error fetching documents from relational database: {e}"
            )
            documents_overview = []

        if documents_overview:
            document_ids_to_purge.update(
                doc.id for doc in documents_overview["results"]
            )

        if not document_ids_to_purge:
            raise R2RException(
                status_code=404, message="No entries found for deletion."
            )

        for document_id in document_ids_to_purge:
            try:
                await self.providers.database.relational.delete_from_documents_overview(
                    str(document_id)
                )
                logger.info(
                    f"Deleted document ID {document_id} from documents_overview."
                )
            except Exception as e:
                logger.error(
                    f"Error deleting document ID {document_id} from documents_overview: {e}"
                )

        return None

    @telemetry_event("DownloadFile")
    async def download_file(
        self, document_id: UUID
    ) -> Optional[Tuple[str, BinaryIO, int]]:
        if result := await self.providers.file.retrieve_file(document_id):
            return result
        return None

    @telemetry_event("DocumentsOverview")
    async def documents_overview(
        self,
        user_ids: Optional[list[UUID]] = None,
        collection_ids: Optional[list[UUID]] = None,
        document_ids: Optional[list[UUID]] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 1000,
        *args: Any,
        **kwargs: Any,
    ):
        return await self.providers.database.relational.get_documents_overview(
            filter_document_ids=document_ids,
            filter_user_ids=user_ids,
            filter_collection_ids=collection_ids,
            offset=offset,
            limit=limit,
        )

    @telemetry_event("DocumentChunks")
    async def document_chunks(
        self,
        document_id: UUID,
        offset: int = 0,
        limit: int = 100,
        *args,
        **kwargs,
    ):
        return self.providers.database.vector.get_document_chunks(
            document_id, offset=offset, limit=limit
        )

    @telemetry_event("InspectKnowledgeGraph")
    async def inspect_knowledge_graph(
        self,
        offset: int = 0,
        limit=1000,
        print_descriptions: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        if self.providers.kg is None:
            raise R2RException(
                status_code=404, message="Knowledge Graph provider not found."
            )

        rel_query = f"""
        MATCH (n1)-[r]->(n2)
        return n1.name AS subject, n1.description AS subject_description, n2.name AS object, n2.description AS object_description, type(r) AS relation, r.description AS relation_description
        SKIP {offset}
        LIMIT {limit}
        """

        try:
            neo4j_results = self.providers.kg.structured_query(
                rel_query
            ).records

            relationships_raw = [
                {
                    "subject": {
                        "name": record["subject"],
                        "description": record["subject_description"],
                    },
                    "relation": {
                        "name": record["relation"],
                        "description": record["relation_description"],
                    },
                    "object": {
                        "name": record["object"],
                        "description": record["object_description"],
                    },
                }
                for record in neo4j_results
            ]

            descriptions_dict = {}
            relationships = []

            for relationship in relationships_raw:
                if print_descriptions:
                    descriptions_dict[relationship["subject"]["name"]] = (
                        relationship["subject"]["description"]
                    )
                    descriptions_dict[relationship["object"]["name"]] = (
                        relationship["object"]["description"]
                    )

                relationships.append(
                    (
                        relationship["subject"]["name"],
                        relationship["relation"]["name"],
                        relationship["object"]["name"],
                    )
                )

            # Create graph representation and group relationships
            graph, grouped_relationships = self._process_relationships(
                relationships
            )

            # Generate output
            output = self.generate_output(
                grouped_relationships,
                graph,
                descriptions_dict,
                print_descriptions,
            )

            return "\n".join(output)

        except Exception as e:
            logger.error("Error printing relationships", exc_info=True)
            raise R2RException(
                status_code=500,
                message=f"An error occurred while fetching relationships: {str(e)}",
            )

    @telemetry_event("AssignDocumentToCollection")
    async def assign_document_to_collection(
        self, document_id: str, collection_id: UUID
    ):

        await self.providers.database.relational.assign_document_to_collection(
            document_id, collection_id
        )
        self.providers.database.vector.assign_document_to_collection(
            document_id, collection_id
        )
        return {"message": "Document assigned to collection successfully"}

    @telemetry_event("RemoveDocumentFromCollection")
    async def remove_document_from_collection(
        self, document_id: str, collection_id: UUID
    ):
        await self.providers.database.relational.remove_document_from_collection(
            document_id, collection_id
        )
        self.providers.database.vector.remove_document_from_collection(
            document_id, collection_id
        )
        return {"message": "Document removed from collection successfully"}

    @telemetry_event("DocumentCollections")
    async def document_collections(
        self, document_id: str, offset: int = 0, limit: int = 100
    ):
        return await self.providers.database.relational.document_collections(
            document_id, offset=offset, limit=limit
        )

    def _process_relationships(
        self, relationships: list[Tuple[str, str, str]]
    ) -> Tuple[Dict[str, list[str]], Dict[str, Dict[str, list[str]]]]:
        graph = defaultdict(list)
        grouped: Dict[str, Dict[str, list[str]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for subject, relation, obj in relationships:
            graph[subject].append(obj)
            grouped[subject][relation].append(obj)
            if obj not in graph:
                graph[obj] = []
        return dict(graph), dict(grouped)

    def generate_output(
        self,
        grouped_relationships: Dict[str, Dict[str, list[str]]],
        graph: Dict[str, list[str]],
        descriptions_dict: Dict[str, str],
        print_descriptions: bool = True,
    ) -> list[str]:
        output = []
        # Print grouped relationships
        for subject, relations in grouped_relationships.items():
            output.append(f"\n== {subject} ==")
            if print_descriptions and subject in descriptions_dict:
                output.append(f"\tDescription: {descriptions_dict[subject]}")
            for relation, objects in relations.items():
                output.append(f"  {relation}:")
                for obj in objects:
                    output.append(f"    - {obj}")
                    if print_descriptions and obj in descriptions_dict:
                        output.append(
                            f"      Description: {descriptions_dict[obj]}"
                        )

        # Print basic graph statistics
        output.extend(
            [
                "\n== Graph Statistics ==",
                f"Number of nodes: {len(graph)}",
                f"Number of edges: {sum(len(neighbors) for neighbors in graph.values())}",
                f"Number of connected components: {self._count_connected_components(graph)}",
            ]
        )

        # Find central nodes
        central_nodes = self._get_central_nodes(graph)
        output.extend(
            [
                "\n== Most Central Nodes ==",
                *(
                    f"  {node}: {centrality:.4f}"
                    for node, centrality in central_nodes
                ),
            ]
        )

        return output

    def _count_connected_components(self, graph: Dict[str, list[str]]) -> int:
        visited = set()
        components = 0

        def dfs(node):
            visited.add(node)
            for neighbor in graph[node]:
                if neighbor not in visited:
                    dfs(neighbor)

        for node in graph:
            if node not in visited:
                dfs(node)
                components += 1

        return components

    def _get_central_nodes(
        self, graph: Dict[str, list[str]]
    ) -> list[Tuple[str, float]]:
        degree = {node: len(neighbors) for node, neighbors in graph.items()}
        total_nodes = len(graph)
        centrality = {
            node: deg / (total_nodes - 1) for node, deg in degree.items()
        }
        return sorted(centrality.items(), key=lambda x: x[1], reverse=True)[:5]

    @telemetry_event("CreateCollection")
    async def create_collection(
        self, name: str, description: str = ""
    ) -> UUID:
        return await self.providers.database.relational.create_collection(
            name, description
        )

    @telemetry_event("GetCollection")
    async def get_collection(self, collection_id: UUID) -> Optional[dict]:
        return await self.providers.database.relational.get_collection(
            collection_id
        )

    @telemetry_event("UpdateCollection")
    async def update_collection(
        self,
        collection_id: UUID,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> bool:
        return await self.providers.database.relational.update_collection(
            collection_id, name, description
        )

    @telemetry_event("DeleteCollection")
    async def delete_collection(self, collection_id: UUID) -> bool:
        await self.providers.database.relational.delete_collection(
            collection_id
        )
        self.providers.database.vector.delete_collection(collection_id)
        return True

    @telemetry_event("ListCollections")
    async def list_collections(
        self, offset: int = 0, limit: int = 100
    ) -> list[dict]:
        return await self.providers.database.relational.list_collections(
            offset=offset, limit=limit
        )

    @telemetry_event("AddUserToCollection")
    async def add_user_to_collection(
        self, user_id: UUID, collection_id: UUID
    ) -> bool:
        return await self.providers.database.relational.add_user_to_collection(
            user_id, collection_id
        )

    @telemetry_event("RemoveUserFromCollection")
    async def remove_user_from_collection(
        self, user_id: UUID, collection_id: UUID
    ) -> bool:
        return await self.providers.database.relational.remove_user_from_collection(
            user_id, collection_id
        )

    @telemetry_event("GetUsersInCollection")
    async def get_users_in_collection(
        self, collection_id: UUID, offset: int = 0, limit: int = 100
    ) -> list[dict]:
        return (
            await self.providers.database.relational.get_users_in_collection(
                collection_id, offset=offset, limit=limit
            )
        )

    @telemetry_event("GetCollectionsForUser")
    async def get_collections_for_user(
        self, user_id: UUID, offset: int = 0, limit: int = 100
    ) -> list[dict]:
        return (
            await self.providers.database.relational.get_collections_for_user(
                user_id, offset, limit
            )
        )

    @telemetry_event("CollectionsOverview")
    async def collections_overview(
        self,
        collection_ids: Optional[list[UUID]] = None,
        offset: int = 0,
        limit: int = 100,
        *args,
        **kwargs,
    ):
        return (
            await self.providers.database.relational.get_collections_overview(
                (
                    [str(ele) for ele in collection_ids]
                    if collection_ids
                    else None
                ),
                offset=offset,
                limit=limit,
            )
        )

    @telemetry_event("GetDocumentsInCollection")
    async def documents_in_collection(
        self, collection_id: UUID, offset: int = 0, limit: int = 100
    ) -> list[dict]:
        return (
            await self.providers.database.relational.documents_in_collection(
                collection_id, offset=offset, limit=limit
            )
        )

    @telemetry_event("AddPrompt")
    async def add_prompt(
        self, name: str, template: str, input_types: dict[str, str]
    ) -> dict:
        try:
            await self.providers.prompt.add_prompt(name, template, input_types)
            return {"message": f"Prompt '{name}' added successfully."}
        except ValueError as e:
            raise R2RException(status_code=400, message=str(e))

    @telemetry_event("GetPrompt")
    async def get_prompt(
        self,
        prompt_name: str,
        inputs: Optional[dict[str, Any]] = None,
        prompt_override: Optional[str] = None,
    ) -> dict:
        try:
            return {
                "message": self.providers.prompt.get_prompt(
                    prompt_name, inputs, prompt_override
                )
            }
        except ValueError as e:
            raise R2RException(status_code=404, message=str(e))

    @telemetry_event("GetAllPrompts")
    async def get_all_prompts(self) -> dict[str, Prompt]:
        return self.providers.prompt.get_all_prompts()

    @telemetry_event("UpdatePrompt")
    async def update_prompt(
        self,
        name: str,
        template: Optional[str] = None,
        input_types: Optional[dict[str, str]] = None,
    ) -> dict:
        try:
            await self.providers.prompt.update_prompt(
                name, template, input_types
            )
            return {"message": f"Prompt '{name}' updated successfully."}
        except ValueError as e:
            raise R2RException(status_code=404, message=str(e))

    @telemetry_event("DeletePrompt")
    async def delete_prompt(self, name: str) -> dict:
        try:
            await self.providers.prompt.delete_prompt(name)
            return {"message": f"Prompt '{name}' deleted successfully."}
        except ValueError as e:
            raise R2RException(status_code=404, message=str(e))



import io
import logging
from typing import BinaryIO, Optional
from uuid import UUID

import asyncpg

from core.base import FileConfig, R2RException
from core.base.providers import FileProvider
from core.providers.database.postgres import PostgresDBProvider

logger = logging.getLogger(__name__)


class PostgresFileProvider(FileProvider):
    def __init__(self, config: FileConfig, db_provider: PostgresDBProvider):
        super().__init__(config)
        self.config: FileConfig = config
        self.db_provider = db_provider
        self.pool = None

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._close_connection()

    async def _close_connection(self):
        if self.pool:
            await self.pool.close()
            self.pool = None

    async def initialize(self):
        self.pool = await asyncpg.create_pool(
            self.db_provider.connection_string
        )
        logger.info(
            "File provider successfully connected to Postgres database."
        )

        async with self.pool.acquire() as conn:
            await conn.execute('CREATE EXTENSION IF NOT EXISTS "lo";')

        await self.create_table()

    def _get_table_name(self, base_name: str) -> str:
        return f"{base_name}"

    async def create_table(self):
        query = f"""
        CREATE TABLE IF NOT EXISTS {self._get_table_name('file_storage')} (
            document_id UUID PRIMARY KEY,
            file_name TEXT NOT NULL,
            file_oid OID NOT NULL,
            file_size BIGINT NOT NULL,
            file_type TEXT,
            created_at TIMESTAMPTZ DEFAULT NOW(),
            updated_at TIMESTAMPTZ DEFAULT NOW()
        );
        """
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(query)

    async def upsert_file(
        self,
        document_id: UUID,
        file_name: str,
        file_oid: int,
        file_size: int,
        file_type: Optional[str] = None,
    ) -> None:
        if not self.pool:
            raise R2RException(
                status_code=500,
                message="Connection to the database is not initialized",
            )

        query = f"""
        INSERT INTO {self._get_table_name('file_storage')}
        (document_id, file_name, file_oid, file_size, file_type)
        VALUES ($1, $2, $3, $4, $5)
        ON CONFLICT (document_id) DO UPDATE SET
            file_name = EXCLUDED.file_name,
            file_oid = EXCLUDED.file_oid,
            file_size = EXCLUDED.file_size,
            file_type = EXCLUDED.file_type,
            updated_at = NOW();
        """
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    query,
                    document_id,
                    file_name,
                    file_oid,
                    file_size,
                    file_type,
                )

    async def store_file(
        self, document_id, file_name, file_content: io.BytesIO, file_type=None
    ):
        if not self.pool:
            raise R2RException(
                status_code=500,
                message="Connection to the database is not initialized",
            )

        file_size = file_content.getbuffer().nbytes
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                oid = await conn.fetchval("SELECT lo_create(0)")
                await self._write_lobject(conn, oid, file_content)
                await self.upsert_file(
                    document_id, file_name, oid, file_size, file_type
                )

    async def _write_lobject(self, conn, oid, file_content):
        # Open the large object
        lobject = await conn.fetchval(
            "SELECT lo_open($1, $2)", oid, 0x20000
        )  # 0x20000 is INV_WRITE flag

        try:
            # Write the content in chunks
            chunk_size = 8192  # 8 KB chunks
            while True:
                chunk = file_content.read(chunk_size)
                if not chunk:
                    break
                await conn.execute("SELECT lowrite($1, $2)", lobject, chunk)

            # Close the large object
            await conn.execute("SELECT lo_close($1)", lobject)

        except Exception as e:
            # Handle exceptions, rollback the transaction if necessary
            await conn.execute(
                "SELECT lo_unlink($1)", oid
            )  # Delete the large object
            raise R2RException(
                status_code=500,
                message=f"Failed to write to large object: {e}",
            )

    async def retrieve_file(
        self, document_id: UUID
    ) -> Optional[tuple[str, BinaryIO, int]]:
        if not self.pool:
            raise R2RException(
                status_code=500,
                message="Connection to the database is not initialized",
            )

        query = f"""
        SELECT file_name, file_oid, file_size
        FROM {self._get_table_name('file_storage')}
        WHERE document_id = $1
        """
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                result = await conn.fetchrow(query, document_id)
                if not result:
                    raise R2RException(
                        status_code=404,
                        message=f"File for document {document_id} not found",
                    )
                file_name, oid, file_size = result
                file_content = await self._read_lobject(conn, oid)
                return file_name, io.BytesIO(file_content), file_size

    async def _read_lobject(self, conn, oid: int) -> bytes:
        file_data = io.BytesIO()
        chunk_size = 8192

        async with conn.transaction():
            try:
                # Check if the large object exists before opening
                lo_exists = await conn.fetchval(
                    "SELECT EXISTS(SELECT 1 FROM pg_largeobject WHERE loid = $1)",
                    oid,
                )
                if not lo_exists:
                    raise R2RException(
                        status_code=404,
                        message=f"Large object {oid} not found.",
                    )

                # Open the large object and fetch its descriptor
                lobject = await conn.fetchval(
                    "SELECT lo_open($1, 262144)", oid
                )  # INV_READ

                # Ensure the descriptor is valid before reading
                if lobject is None:
                    raise R2RException(
                        status_code=404,
                        message=f"Failed to open large object {oid}.",
                    )

                # Read the large object in chunks
                while True:
                    chunk = await conn.fetchval(
                        "SELECT loread($1, $2)", lobject, chunk_size
                    )
                    if not chunk:
                        break
                    file_data.write(chunk)
            except asyncpg.exceptions.UndefinedObjectError as e:
                # Handle large object not found
                raise R2RException(
                    status_code=404,
                    message=f"Failed to read large object {oid}: {e}",
                )
            finally:
                # Always close the large object
                await conn.execute("SELECT lo_close($1)", lobject)

        return file_data.getvalue()

    async def delete_file(self, document_id: UUID) -> bool:
        if not self.pool:
            raise R2RException(
                status_code=500,
                message="Connection to the database is not initialized",
            )

        query = f"""
        SELECT file_oid FROM {self._get_table_name('file_storage')}
        WHERE document_id = $1
        """
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                result = await conn.fetchval(query, document_id)
                if not result:
                    raise R2RException(
                        status_code=404,
                        message=f"File for document {document_id} not found",
                    )
                oid = result
                await self._delete_lobject(conn, oid)

                delete_query = f"""
                DELETE FROM {self._get_table_name('file_storage')}
                WHERE document_id = $1
                """
                await conn.execute(delete_query, document_id)
        return True

    async def _delete_lobject(self, conn, oid: int) -> None:
        await conn.execute("SELECT lo_unlink($1)", oid)

    # Implementation of get_files_overview
    async def get_files_overview(
        self,
        filter_document_ids: Optional[list[UUID]] = None,
        filter_file_names: Optional[list[str]] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> list[dict]:
        if not self.pool:
            raise R2RException(
                status_code=500,
                message="Connection to the database is not initialized",
            )

        conditions = []
        params = []
        query = f"""
        SELECT document_id, file_name, file_oid, file_size, file_type, created_at, updated_at
        FROM {self._get_table_name('file_storage')}
        """

        if filter_document_ids:
            conditions.append(f"document_id = ANY(${len(params) + 1})")
            params.append([str(doc_id) for doc_id in filter_document_ids])

        if filter_file_names:
            conditions.append(f"file_name = ANY(${len(params) + 1})")
            params.append(filter_file_names)

        if conditions:
            query += " WHERE " + " AND ".join(conditions)

        query += f" ORDER BY created_at DESC OFFSET ${len(params) + 1} LIMIT ${len(params) + 2}"
        params.extend([offset, limit])

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                results = await conn.fetch(query, *params)

        if not results:
            raise R2RException(
                status_code=404,
                message="No files found with the given filters",
            )

        return [
            {
                "document_id": row["document_id"],
                "file_name": row["file_name"],
                "file_oid": row["file_oid"],
                "file_size": row["file_size"],
                "file_type": row["file_type"],
                "created_at": row["created_at"],
                "updated_at": row["updated_at"],
            }
            for row in results
        ]



import asyncio
import logging

from hatchet_sdk import Context

from core.base import IngestionStatus, increment_version
from core.base.abstractions import DocumentInfo, R2RException

from ..services import IngestionService, IngestionServiceAdapter
from .base import r2r_hatchet

logger = logging.getLogger(__name__)


@r2r_hatchet.workflow(
    name="ingest-file",
    timeout="60m",
)
class IngestFilesWorkflow:
    def __init__(self, ingestion_service: IngestionService):
        self.ingestion_service = ingestion_service

    @r2r_hatchet.step(timeout="60m")
    async def parse(self, context: Context) -> dict:
        input_data = context.workflow_input()["request"]
        parsed_data = IngestionServiceAdapter.parse_ingest_file_input(
            input_data
        )

        ingestion_result = await self.ingestion_service.ingest_file_ingress(
            **parsed_data
        )

        document_info = ingestion_result["info"]

        await self.ingestion_service.update_document_status(
            document_info,
            status=IngestionStatus.PARSING,
        )

        extractions_generator = await self.ingestion_service.parse_file(
            document_info
        )

        extractions = []
        async for extraction in extractions_generator:
            extractions.append(extraction)

        serializable_extractions = [
            fragment.to_dict() for fragment in extractions
        ]

        return {
            "status": "Successfully extracted data",
            "extractions": serializable_extractions,
            "document_info": document_info.to_dict(),
        }

    @r2r_hatchet.step(parents=["parse"], timeout="60m")
    async def chunk(self, context: Context) -> dict:
        document_info_dict = context.step_output("parse")["document_info"]
        document_info = DocumentInfo(**document_info_dict)

        await self.ingestion_service.update_document_status(
            document_info,
            status=IngestionStatus.CHUNKING,
        )

        extractions = context.step_output("parse")["extractions"]
        chunking_config = context.workflow_input()["request"].get(
            "chunking_config"
        )

        chunk_generator = await self.ingestion_service.chunk_document(
            extractions,
            chunking_config,
        )

        chunks = []
        async for chunk in chunk_generator:
            chunks.append(chunk)

        serializable_chunks = [chunk.to_dict() for chunk in chunks]

        return {
            "status": "Successfully chunked data",
            "chunks": serializable_chunks,
            "document_info": document_info.to_dict(),
        }

    @r2r_hatchet.step(parents=["chunk"], timeout="60m")
    async def embed(self, context: Context) -> dict:
        document_info_dict = context.step_output("chunk")["document_info"]
        document_info = DocumentInfo(**document_info_dict)

        await self.ingestion_service.update_document_status(
            document_info,
            status=IngestionStatus.EMBEDDING,
        )

        chunks = context.step_output("chunk")["chunks"]

        embedding_generator = await self.ingestion_service.embed_document(
            chunks
        )

        embeddings = []
        async for embedding in embedding_generator:
            embeddings.append(embedding)

        await self.ingestion_service.update_document_status(
            document_info,
            status=IngestionStatus.STORING,
        )

        storage_generator = await self.ingestion_service.store_embeddings(  # type: ignore
            embeddings
        )

        async for _ in storage_generator:
            pass

        return {
            "document_info": document_info.to_dict(),
        }

    @r2r_hatchet.step(parents=["embed"], timeout="60m")
    async def finalize(self, context: Context) -> dict:
        document_info_dict = context.step_output("embed")["document_info"]
        document_info = DocumentInfo(**document_info_dict)

        is_update = context.workflow_input()["request"].get("is_update")

        await self.ingestion_service.finalize_ingestion(
            document_info, is_update=is_update
        )

        await self.ingestion_service.update_document_status(
            document_info,
            status=IngestionStatus.SUCCESS,
        )

        return {
            "status": "Successfully finalized ingestion",
            "document_info": document_info.to_dict(),
        }

    @r2r_hatchet.on_failure_step()
    async def on_failure(self, context: Context) -> None:
        request = context.workflow_input().get("request", {})
        document_id = request.get("document_id")

        if not document_id:
            logger.error(
                "No document id was found in workflow input to mark a failure."
            )
            return

        try:
            documents_overview = await self.ingestion_service.providers.database.relational.get_documents_overview(
                filter_document_ids=[document_id]
            )

            if not documents_overview:
                logger.error(
                    f"Document with id {document_id} not found in database to mark failure."
                )
                return

            document_info = documents_overview[0]

            # Update the document status to FAILURE
            await self.ingestion_service.update_document_status(
                document_info,
                status=IngestionStatus.FAILURE,
            )

        except Exception as e:
            logger.error(
                f"Failed to update document status for {document_id}: {e}"
            )


# TODO: Implement a check to see if the file is actually changed before updating
@r2r_hatchet.workflow(name="update-files", timeout="60m")
class UpdateFilesWorkflow:
    def __init__(self, ingestion_service: IngestionService):
        self.ingestion_service = ingestion_service

    @r2r_hatchet.step(retries=0, timeout="60m")
    async def update_files(self, context: Context) -> None:
        data = context.workflow_input()["request"]
        parsed_data = IngestionServiceAdapter.parse_update_files_input(data)

        file_datas = parsed_data["file_datas"]
        user = parsed_data["user"]
        document_ids = parsed_data["document_ids"]
        metadatas = parsed_data["metadatas"]
        chunking_config = parsed_data["chunking_config"]
        file_sizes_in_bytes = parsed_data["file_sizes_in_bytes"]

        if not file_datas:
            raise R2RException(
                status_code=400, message="No files provided for update."
            )
        if len(document_ids) != len(file_datas):
            raise R2RException(
                status_code=400,
                message="Number of ids does not match number of files.",
            )

        documents_overview = (
            await self.ingestion_service.providers.database.relational.get_documents_overview(
                filter_document_ids=document_ids,
                filter_user_ids=None if user.is_superuser else [user.id],
            )
        )["results"]
        if len(documents_overview) != len(document_ids):
            raise R2RException(
                status_code=404,
                message="One or more documents not found.",
            )

        results = []

        for idx, (
            file_data,
            doc_id,
            doc_info,
            file_size_in_bytes,
        ) in enumerate(
            zip(
                file_datas,
                document_ids,
                documents_overview,
                file_sizes_in_bytes,
            )
        ):
            new_version = increment_version(doc_info.version)

            updated_metadata = (
                metadatas[idx] if metadatas else doc_info.metadata
            )
            updated_metadata["title"] = (
                updated_metadata.get("title")
                or file_data["filename"].split("/")[-1]
            )

            # Prepare input for ingest_file workflow
            ingest_input = {
                "file_data": file_data,
                "user": data.get("user"),
                "metadata": updated_metadata,
                "document_id": str(doc_id),
                "version": new_version,
                "chunking_config": (
                    chunking_config.model_dump_json()
                    if chunking_config
                    else None
                ),
                "size_in_bytes": file_size_in_bytes,
                "is_update": True,
            }

            # Spawn ingest_file workflow as a child workflow
            child_result = (
                await context.aio.spawn_workflow(
                    "ingest-file",
                    {"request": ingest_input},
                    key=f"ingest_file_{doc_id}",
                )
            ).result()
            results.append(child_result)

        await asyncio.gather(*results)

        return None



from .abstractions import *
from .agent import *
from .api.models import *
from .logging import *
from .parsers import *
from .pipeline import *
from .pipes import *
from .providers import *
from .utils import *

__all__ = [
    ## ABSTRACTIONS
    # Base abstractions
    "AsyncSyncMeta",
    "syncable",
    # Completion abstractions
    "CompletionRecord",
    "MessageType",
    # Document abstractions
    "DataType",
    "Document",
    "DocumentExtraction",
    "DocumentFragment",
    "DocumentInfo",
    "IngestionStatus",
    "RestructureStatus",
    "DocumentType",
    # Embedding abstractions
    "EmbeddingPurpose",
    "default_embedding_prefixes",
    # Exception abstractions
    "R2RDocumentProcessingError",
    "R2RException",
    # KG abstractions
    "Entity",
    "KGExtraction",
    "Triple",
    # LLM abstractions
    "GenerationConfig",
    "LLMChatCompletion",
    "LLMChatCompletionChunk",
    "RAGCompletion",
    # Prompt abstractions
    "Prompt",
    # Search abstractions
    "AggregateSearchResult",
    "KGSearchResult",
    "KGSearchSettings",
    "VectorSearchResult",
    "VectorSearchSettings",
    # Restructure abstractions
    "KGCreationSettings",
    "KGEnrichmentSettings",
    # User abstractions
    "Token",
    "TokenData",
    "UserStats",
    # Vector abstractions
    "Vector",
    "VectorEntry",
    "VectorType",
    "StorageResult",
    ## AGENT
    # Agent abstractions
    "Agent",
    "AgentConfig",
    "Conversation",
    "Message",
    "Tool",
    "ToolResult",
    ## API
    # Auth Responses
    "GenericMessageResponse",
    "TokenResponse",
    "UserResponse",
    ## LOGGING
    # Basic types
    "RunType",
    "AnalysisTypes",
    "LogAnalytics",
    "LogAnalyticsConfig",
    "LogFilterCriteria",
    "LogProcessor",
    # Logging Providers
    "LocalRunLoggingProvider",
    "LoggingConfig",
    "PostgresLoggingConfig",
    "PostgresRunLoggingProvider",
    "RunLoggingSingleton",
    # Run Manager
    "RunManager",
    "manage_run",
    ## PARSERS
    # Base parser
    "AsyncParser",
    ## PIPELINE
    # Base pipeline
    "AsyncPipeline",
    ## PIPES
    "AsyncPipe",
    "AsyncState",
    "PipeType",
    ## PROVIDERS
    # Base provider classes
    "Provider",
    "ProviderConfig",
    # Auth provider
    "AuthConfig",
    "AuthProvider",
    # Chunking provider
    "R2RChunkingConfig",
    "UnstructuredChunkingConfig",
    "ChunkingConfig",
    "ChunkingProvider",
    "Strategy",
    # Crypto provider
    "CryptoConfig",
    "CryptoProvider",
    # Database providers
    "DatabaseConfig",
    "DatabaseProvider",
    "RelationalDBProvider",
    "VectorDBProvider",
    # Embedding provider
    "EmbeddingConfig",
    "EmbeddingProvider",
    # File provider
    "FileConfig",
    "FileProvider",
    # Knowledge Graph provider
    "KGConfig",
    "KGProvider",
    # LLM provider
    "CompletionConfig",
    "CompletionProvider",
    # Parsing provider
    "ParsingConfig",
    "ParsingProvider",
    "OverrideParser",
    # Prompt provider
    "PromptConfig",
    "PromptProvider",
    ## UTILS
    "RecursiveCharacterTextSplitter",
    "TextSplitter",
    "run_pipeline",
    "to_async_generator",
    "format_search_results_for_llm",
    "format_search_results_for_stream",
    "generate_run_id",
    "generate_id_from_label",
    "increment_version",
    "EntityType",
    "RelationshipType",
    "format_entity_types",
    "format_relations",
]



import asyncio
import base64
import concurrent.futures
import logging
import os
from io import BytesIO
from typing import Dict, List

from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from unstructured.partition.auto import partition

logger = logging.getLogger(__name__)

app = FastAPI()


class PartitionRequestModel(BaseModel):
    file_content: str
    chunking_config: Dict


class PartitionResponseModel(BaseModel):
    elements: List[Dict]


executor = concurrent.futures.ThreadPoolExecutor(
    max_workers=int(os.environ.get("MAX_INGESTION_WORKERS", 10))
)


def run_partition(file_content: str, chunking_config: Dict) -> List[Dict]:
    file_content_bytes = base64.b64decode(file_content)
    file_io = BytesIO(file_content_bytes)
    elements = partition(file=file_io, **chunking_config)
    return [element.to_dict() for element in elements]


@app.get("/health")
async def health_endpoint():
    return {"status": "ok"}


@app.post("/partition", response_model=PartitionResponseModel)
async def partition_endpoint(request: PartitionRequestModel):
    try:
        logger.info(f"Partitioning request received")
        loop = asyncio.get_event_loop()
        elements = await loop.run_in_executor(
            executor,
            run_partition,
            request.file_content,
            request.chunking_config,
        )
        logger.info(f"Partitioning completed")
        return PartitionResponseModel(elements=elements)
    except Exception as e:
        logger.error(f"Error partitioning file: {str(e)}")
        raise HTTPException(status_code=500, detail=str(e))



import asyncio
import contextvars
from contextlib import asynccontextmanager
from typing import Optional
from uuid import UUID

from core.base.api.models import UserResponse
from core.base.logging.base import RunType
from core.base.utils import generate_run_id

from .run_logger import RunLoggingSingleton

run_id_var = contextvars.ContextVar("run_id", default=generate_run_id())


class RunManager:
    def __init__(self, logger: RunLoggingSingleton):
        self.logger = logger
        self.run_info: dict[UUID, dict] = {}

    async def set_run_info(self, run_type: str, run_id: Optional[UUID] = None):
        run_id = run_id or run_id_var.get()
        if run_id is None:
            run_id = generate_run_id()
            token = run_id_var.set(run_id)
            self.run_info[run_id] = {"run_type": run_type}
        else:
            token = run_id_var.set(run_id)
        return run_id, token

    async def get_info_logs(self):
        run_id = run_id_var.get()
        return self.run_info.get(run_id, None)

    async def log_run_info(
        self,
        run_type: RunType,
        user: UserResponse,
    ):
        if asyncio.iscoroutine(user):
            user = await user

        if run_id := run_id_var.get():
            await self.logger.info_log(
                run_id=run_id,
                run_type=run_type,
                user_id=user.id,
            )
        else:
            raise ValueError("No run ID set")

    async def clear_run_info(self, token: contextvars.Token):
        run_id = run_id_var.get()
        run_id_var.reset(token)
        if run_id and run_id in self.run_info:
            del self.run_info[run_id]


@asynccontextmanager
async def manage_run(
    run_manager: RunManager,
    run_type: RunType = RunType.UNSPECIFIED,
    run_id: Optional[UUID] = None,
):
    run_id, token = await run_manager.set_run_info(run_type, run_id)
    try:
        yield run_id
    finally:
        # Check if we're in a test environment
        if isinstance(token, contextvars.Token):
            run_id_var.reset(token)
        else:
            # We're in a test environment, just reset the run_id_var
            run_id_var.set(None)  # type: ignore



import json
import os
from contextlib import ExitStack
from typing import Optional, Union
from uuid import UUID

from .models import ChunkingConfig


class IngestionMethods:

    @staticmethod
    async def ingest_files(
        client,
        file_paths: list[str],
        document_ids: Optional[list[Union[str, UUID]]] = None,
        metadatas: Optional[list[dict]] = None,
        chunking_config: Optional[Union[dict, ChunkingConfig]] = None,
    ) -> dict:
        """
        Ingest files into your R2R deployment

        Args:
            file_paths (List[str]): List of file paths to ingest.
            document_ids (Optional[List[str]]): List of document IDs.
            metadatas (Optional[List[dict]]): List of metadata dictionaries for each file.
            chunking_config (Optional[Union[dict, ChunkingConfig]]): Custom chunking configuration.

        Returns:
            dict: Ingestion results containing processed, failed, and skipped documents.
        """
        if document_ids is not None and len(file_paths) != len(document_ids):
            raise ValueError(
                "Number of file paths must match number of document IDs."
            )
        if metadatas is not None and len(file_paths) != len(metadatas):
            raise ValueError(
                "Number of metadatas must match number of document IDs."
            )
        if chunking_config is not None and not isinstance(
            chunking_config, ChunkingConfig
        ):
            # check if the provided dict maps to a ChunkingConfig
            ChunkingConfig(**chunking_config)

        all_file_paths: list[str] = []
        for path in file_paths:
            if os.path.isdir(path):
                for root, _, files in os.walk(path):
                    all_file_paths.extend(
                        os.path.join(root, file) for file in files
                    )
            else:
                all_file_paths.append(path)

        with ExitStack() as stack:
            files_tuples = [
                (
                    "files",
                    (
                        os.path.basename(file),
                        stack.enter_context(open(file, "rb")),
                        "application/octet-stream",
                    ),
                )
                for file in all_file_paths
            ]

            data = {}
            if document_ids:
                data["document_ids"] = json.dumps(
                    [str(doc_id) for doc_id in document_ids]
                )
            if metadatas:
                data["metadatas"] = json.dumps(metadatas)

            if chunking_config:
                data["chunking_config"] = (
                    chunking_config.model_dump()  # type: ignore
                    if isinstance(chunking_config, ChunkingConfig)
                    else chunking_config
                )

            return await client._make_request(
                "POST", "ingest_files", data=data, files=files_tuples
            )

    @staticmethod
    async def retry_ingest_files(
        client,
        document_ids: list[Union[str, UUID]],
    ) -> dict:
        """
        Retry ingestion for failed documents.

        Args:
            document_ids (List[Union[str, UUID]]): List of document IDs to retry.

        Returns:
            dict: Retry results containing processed, failed, and skipped documents.
        """
        return await client._make_request(
            "POST",
            "retry_ingest_files",
            data={"document_ids": [str(doc_id) for doc_id in document_ids]},
        )

    @staticmethod
    async def update_files(
        client,
        file_paths: list[str],
        document_ids: Optional[list[Union[str, UUID]]] = None,
        metadatas: Optional[list[dict]] = None,
        chunking_config: Optional[Union[dict, ChunkingConfig]] = None,
    ) -> dict:
        """
        Update existing files in your R2R deployment.

        Args:
            file_paths (List[str]): List of file paths to update.
            document_ids (List[str]): List of document IDs to update.
            metadatas (Optional[List[dict]]): List of updated metadata dictionaries for each file.
            chunking_config (Optional[Union[dict, ChunkingConfig]]): Custom chunking configuration.

        Returns:
            dict: Update results containing processed, failed, and skipped documents.
        """
        if document_ids is not None and len(file_paths) != len(document_ids):
            raise ValueError(
                "Number of file paths must match number of document IDs."
            )
        if metadatas is not None and len(file_paths) != len(metadatas):
            raise ValueError(
                "Number of file paths must match number of document IDs."
            )

        with ExitStack() as stack:
            files = [
                (
                    "files",
                    (
                        os.path.basename(file),
                        stack.enter_context(open(file, "rb")),
                        "application/octet-stream",
                    ),
                )
                for file in file_paths
            ]

            data = {}
            if document_ids:
                data["document_ids"] = json.dumps(
                    [str(doc_id) for doc_id in document_ids]
                )
            if metadatas:
                data["metadatas"] = json.dumps(metadatas)
            if chunking_config:
                data["chunking_config"] = (
                    chunking_config.model_dump()  # type: ignore
                    if isinstance(chunking_config, ChunkingConfig)
                    else chunking_config
                )

            return await client._make_request(
                "POST", "update_files", data=data, files=files
            )



# TODO - Cleanup the handling for non-auth configurations
import json
import mimetypes
from datetime import datetime, timezone
from typing import Optional
from uuid import UUID

import psutil
from fastapi import Body, Depends, Path, Query
from fastapi.responses import StreamingResponse
from pydantic import Json

from core.base import R2RException
from core.base.api.models import (
    WrappedAddUserResponse,
    WrappedAnalyticsResponse,
    WrappedAppSettingsResponse,
    WrappedCollectionListResponse,
    WrappedCollectionOverviewResponse,
    WrappedCollectionResponse,
    WrappedDocumentChunkResponse,
    WrappedDocumentOverviewResponse,
    WrappedGetPromptsResponse,
    WrappedKnowledgeGraphResponse,
    WrappedLogResponse,
    WrappedPromptMessageResponse,
    WrappedScoreCompletionResponse,
    WrappedServerStatsResponse,
    WrappedUserCollectionResponse,
    WrappedUserOverviewResponse,
    WrappedUsersInCollectionResponse,
)
from core.base.logging import AnalysisTypes, LogFilterCriteria
from core.base.providers import OrchestrationProvider

from ..services.management_service import ManagementService
from .base_router import BaseRouter, RunType


class ManagementRouter(BaseRouter):
    def __init__(
        self,
        service: ManagementService,
        run_type: RunType = RunType.MANAGEMENT,
        orchestration_provider: Optional[OrchestrationProvider] = None,
    ):
        super().__init__(service, run_type, orchestration_provider)
        self.service: ManagementService = service  # for type hinting
        self.start_time = datetime.now(timezone.utc)

    def _register_workflows(self):
        pass

    def _load_openapi_extras(self):
        return {}

    # TODO: remove this from the management route, it should be at the base of the server
    def _setup_routes(self):
        @self.router.get("/health")
        @self.base_endpoint
        async def health_check():
            return {"response": "ok"}

        @self.router.get("/server_stats")
        @self.base_endpoint
        async def server_stats(
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedServerStatsResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only an authorized user can call the `server_stats` endpoint.",
                    403,
                )
            return {
                "start_time": self.start_time.isoformat(),
                "uptime_seconds": (
                    datetime.now(timezone.utc) - self.start_time
                ).total_seconds(),
                "cpu_usage": psutil.cpu_percent(),
                "memory_usage": psutil.virtual_memory().percent,
            }

        @self.router.post("/update_prompt")
        @self.base_endpoint
        async def update_prompt_app(
            name: str = Body(..., description="Prompt name"),
            template: Optional[str] = Body(
                None, description="Prompt template"
            ),
            input_types: Optional[dict[str, str]] = Body(
                {}, description="Input types"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedPromptMessageResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `update_prompt` endpoint.",
                    403,
                )

            result = await self.service.update_prompt(
                name, template, input_types
            )
            return result

        @self.router.post("/add_prompt")
        @self.base_endpoint
        async def add_prompt_app(
            name: str = Body(..., description="Prompt name"),
            template: str = Body(..., description="Prompt template"),
            input_types: dict[str, str] = Body({}, description="Input types"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedPromptMessageResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `add_prompt` endpoint.",
                    403,
                )
            result = await self.service.add_prompt(name, template, input_types)
            return result

        @self.router.get("/get_prompt/{prompt_name}")
        @self.base_endpoint
        async def get_prompt_app(
            prompt_name: str = Path(..., description="Prompt name"),
            inputs: Optional[Json[dict]] = Query(
                None, description="JSON-encoded prompt inputs"
            ),
            prompt_override: Optional[str] = Query(
                None, description="Prompt override"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedPromptMessageResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `get_prompt` endpoint.",
                    403,
                )
            result = await self.service.get_prompt(
                prompt_name, inputs, prompt_override
            )
            return result

        @self.router.get("/get_all_prompts")
        @self.base_endpoint
        async def get_all_prompts_app(
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedGetPromptsResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `get_all_prompts` endpoint.",
                    403,
                )
            result = await self.service.get_all_prompts()
            return {"prompts": result}

        @self.router.delete("/delete_prompt/{prompt_name}")
        @self.base_endpoint
        async def delete_prompt_app(
            prompt_name: str = Path(..., description="Prompt name"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=None,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `delete_prompt` endpoint.",
                    403,
                )
            await self.service.delete_prompt(prompt_name)
            return None

        @self.router.get("/analytics")
        @self.base_endpoint
        async def get_analytics_app(
            filter_criteria: Optional[Json[dict]] = Query({}),
            analysis_types: Optional[Json[dict]] = Query({}),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedAnalyticsResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `analytics` endpoint.", 403
                )

            try:
                result = await self.service.analytics(
                    filter_criteria=LogFilterCriteria(filters=filter_criteria),
                    analysis_types=AnalysisTypes(
                        analysis_types=analysis_types
                    ),
                )
                return result
            except json.JSONDecodeError as e:
                raise R2RException(
                    f"Invalid JSON in query parameters: {str(e)}", 400
                )
            except ValueError as e:
                raise R2RException(
                    f"Invalid data in query parameters: {str(e)}", 400
                )

        # TODO: should we add a message to the response model with warnings i.e. if the max_runs passed in violates the max_runs limit?
        @self.router.get("/logs")
        @self.base_endpoint
        async def logs_app(
            run_type_filter: Optional[str] = Query(""),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedLogResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `logs` endpoint.", 403
                )

            return await self.service.logs(
                run_type_filter=run_type_filter,
                offset=offset,
                limit=limit,
            )

        @self.router.get("/app_settings")
        @self.base_endpoint
        async def app_settings(
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedAppSettingsResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `app_settings` endpoint.",
                    403,
                )
            return await self.service.app_settings()

        @self.router.post("/score_completion")
        @self.base_endpoint
        async def score_completion(
            message_id: str = Body(..., description="Message ID"),
            score: float = Body(..., description="Completion score"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedScoreCompletionResponse,
        ):
            message_uuid = UUID(message_id)
            return await self.service.score_completion(
                message_id=message_uuid, score=score
            )

        @self.router.get("/users_overview")
        @self.base_endpoint
        async def users_overview_app(
            user_ids: Optional[list[str]] = Query([]),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedUserOverviewResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `users_overview` endpoint.",
                    403,
                )

            user_uuids = (
                [UUID(user_id) for user_id in user_ids] if user_ids else None
            )

            users_overview_response = await self.service.users_overview(
                user_ids=user_uuids, offset=offset, limit=limit
            )

            return users_overview_response["results"], {
                "total_entries": users_overview_response["total_entries"]
            }

        @self.router.delete("/delete", status_code=204)
        @self.base_endpoint
        async def delete_app(
            filters: str = Query(..., description="JSON-encoded filters"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=None,
        ):
            filters_dict = json.loads(filters) if filters else None
            return await self.service.delete(filters=filters_dict)

        @self.router.get(
            "/download_file/{document_id}", response_class=StreamingResponse
        )
        @self.base_endpoint
        async def download_file_app(
            document_id: str = Path(..., description="Document ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            """
            Download a file by its document ID as a stream.
            """
            # TODO: Add a check to see if the user has access to the file

            try:
                document_uuid = UUID(document_id)
            except ValueError:
                raise R2RException(
                    status_code=400, message="Invalid document ID format."
                )

            file_tuple = await self.service.download_file(document_uuid)
            if not file_tuple:
                raise R2RException(status_code=404, message="File not found.")

            file_name, file_content, file_size = file_tuple

            mime_type, _ = mimetypes.guess_type(file_name)
            if not mime_type:
                mime_type = "application/octet-stream"

            async def file_stream():
                chunk_size = 1024 * 1024  # 1MB
                while True:
                    data = file_content.read(chunk_size)
                    if not data:
                        break
                    yield data

            return StreamingResponse(
                file_stream(),
                media_type=mime_type,
                headers={
                    "Content-Disposition": f'inline; filename="{file_name}"',
                    "Content-Length": str(file_size),
                },
            )

        @self.router.get("/documents_overview")
        @self.base_endpoint
        async def documents_overview_app(
            document_ids: list[str] = Query([]),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedDocumentOverviewResponse,
        ):
            request_user_ids = (
                None if auth_user.is_superuser else [auth_user.id]
            )
            document_uuids = [
                UUID(document_id) for document_id in document_ids
            ]
            documents_overview_response = (
                await self.service.documents_overview(
                    user_ids=request_user_ids,
                    collection_ids=auth_user.collection_ids,
                    document_ids=document_uuids,
                    offset=offset,
                    limit=limit,
                )
            )
            return documents_overview_response["results"], {
                "total_entries": documents_overview_response["total_entries"]
            }

        @self.router.get("/document_chunks/{document_id}")
        @self.base_endpoint
        async def document_chunks_app(
            document_id: str = Path(...),
            offset: Optional[int] = Query(0, ge=0),
            limit: Optional[int] = Query(100, ge=0),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedDocumentChunkResponse,
        ):
            document_uuid = UUID(document_id)

            document_chunks_result = await self.service.document_chunks(
                document_uuid, offset, limit
            )

            if not document_chunks_result:
                raise R2RException(
                    "No chunks found for the given document ID.",
                    404,
                )

            is_owner = str(
                document_chunks_result["results"][0].get("user_id")
            ) == str(auth_user.id)

            if not is_owner and not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can arbitrarily call document_chunks.",
                    403,
                )

            return document_chunks_result["results"], {
                "total_entries": document_chunks_result["total_entries"]
            }

        @self.router.get("/inspect_knowledge_graph")
        @self.base_endpoint
        async def inspect_knowledge_graph(
            offset: int = 0,
            limit: int = 100,
            print_descriptions: bool = False,
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedKnowledgeGraphResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `inspect_knowledge_graph` endpoint.",
                    403,
                )
            return await self.service.inspect_knowledge_graph(
                offset=offset,
                limit=limit,
                print_descriptions=print_descriptions,
            )

        @self.router.get("/collections_overview")
        @self.base_endpoint
        async def collections_overview_app(
            collection_ids: Optional[list[str]] = Query(None),
            offset: Optional[int] = Query(0, ge=0),
            limit: Optional[int] = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedCollectionOverviewResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `collections_overview` endpoint.",
                    403,
                )

            collection_uuids = (
                [UUID(collection_id) for collection_id in collection_ids]
                if collection_ids
                else None
            )
            collections_overview_response = (
                await self.service.collections_overview(
                    collection_ids=collection_uuids, offset=offset, limit=limit
                )
            )

            return collections_overview_response["results"], {
                "total_entries": collections_overview_response["total_entries"]
            }

        @self.router.post("/create_collection")
        @self.base_endpoint
        async def create_collection_app(
            name: str = Body(..., description="Collection name"),
            description: Optional[str] = Body(
                "", description="Collection description"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedCollectionResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can create collections.", 403
                )
            return await self.service.create_collection(name, description)

        @self.router.get("/get_collection/{collection_id}")
        @self.base_endpoint
        async def get_collection_app(
            collection_id: str = Path(..., description="Collection ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedCollectionResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get Collection details.", 403
                )
            collection_uuid = UUID(collection_id)
            result = await self.service.get_collection(collection_uuid)
            return result

        @self.router.put("/update_collection")
        @self.base_endpoint
        async def update_collection_app(
            collection_id: str = Body(..., description="Collection ID"),
            name: Optional[str] = Body(
                None, description="Updated collection name"
            ),
            description: Optional[str] = Body(
                None, description="Updated collection description"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedCollectionResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can update collections.", 403
                )
            collection_uuid = UUID(collection_id)
            return await self.service.update_collection(
                collection_uuid, name, description
            )

        @self.router.delete("/delete_collection/{collection_id}")
        @self.base_endpoint
        async def delete_collection_app(
            collection_id: str = Path(..., description="Collection ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can delete collections.", 403
                )
            collection_uuid = UUID(collection_id)
            return await self.service.delete_collection(collection_uuid)

        @self.router.get("/list_collections")
        @self.base_endpoint
        async def list_collections_app(
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedCollectionListResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can list all collections.", 403
                )
            list_collections_response = await self.service.list_collections(
                offset=offset, limit=min(max(limit, 1), 1000)
            )

            return list_collections_response["results"], {
                "total_entries": list_collections_response["total_entries"]
            }

        @self.router.post("/add_user_to_collection")
        @self.base_endpoint
        async def add_user_to_collection_app(
            user_id: str = Body(..., description="User ID"),
            collection_id: str = Body(..., description="Collection ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedAddUserResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can add users to collections.", 403
                )
            user_uuid = UUID(user_id)
            collection_uuid = UUID(collection_id)
            result = await self.service.add_user_to_collection(
                user_uuid, collection_uuid
            )
            return {"result": result}

        @self.router.post("/remove_user_from_collection")
        @self.base_endpoint
        async def remove_user_from_collection_app(
            user_id: str = Body(..., description="User ID"),
            collection_id: str = Body(..., description="Collection ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can remove users from collections.", 403
                )
            user_uuid = UUID(user_id)
            collection_uuid = UUID(collection_id)
            await self.service.remove_user_from_collection(
                user_uuid, collection_uuid
            )
            return None

        # TODO - Proivde response model
        @self.router.get("/get_users_in_collection/{collection_id}")
        @self.base_endpoint
        async def get_users_in_collection_app(
            collection_id: str = Path(..., description="Collection ID"),
            offset: int = Query(0, ge=0, description="Pagination offset"),
            limit: int = Query(
                100, ge=1, le=1000, description="Pagination limit"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedUsersInCollectionResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get users in a collection.", 403
                )
            collection_uuid = UUID(collection_id)
            users_in_collection_response = (
                await self.service.get_users_in_collection(
                    collection_id=collection_uuid,
                    offset=offset,
                    limit=min(max(limit, 1), 1000),
                )
            )

            return users_in_collection_response["results"], {
                "total_entries": users_in_collection_response["total_entries"]
            }

        @self.router.get("/user_collections/{user_id}")
        @self.base_endpoint
        async def get_collections_for_user_app(
            user_id: str = Path(..., description="User ID"),
            offset: int = Query(0, ge=0, description="Pagination offset"),
            limit: int = Query(
                100, ge=1, le=1000, description="Pagination limit"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedUserCollectionResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get collections for a user.", 403
                )
            user_uuid = UUID(user_id)
            user_collection_response = (
                await self.service.get_collections_for_user(
                    user_uuid, offset, limit
                )
            )

            return user_collection_response["results"], {
                "total_entries": user_collection_response["total_entries"]
            }

        @self.router.post("/assign_document_to_collection")
        @self.base_endpoint
        async def assign_document_to_collection_app(
            document_id: str = Body(..., description="Document ID"),
            collection_id: str = Body(..., description="Collection ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can assign documents to collections.",
                    403,
                )
            document_uuid = UUID(document_id)
            collection_uuid = UUID(collection_id)
            return await self.service.assign_document_to_collection(
                document_uuid, collection_uuid
            )

        @self.router.post("/remove_document_from_collection")
        @self.base_endpoint
        async def remove_document_from_collection_app(
            document_id: str = Body(..., description="Document ID"),
            collection_id: str = Body(..., description="Collection ID"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=None,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can remove documents from collections.",
                    403,
                )
            document_uuid = UUID(document_id)
            collection_uuid = UUID(collection_id)
            await self.service.remove_document_from_collection(
                document_uuid, collection_uuid
            )
            return None

        @self.router.get("/document_collections/{document_id}")
        @self.base_endpoint
        async def document_collections_app(
            document_id: str = Path(..., description="Document ID"),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedCollectionListResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get the collections belonging to a document.",
                    403,
                )
            document_collections_response = (
                await self.service.document_collections(
                    document_id, offset, limit
                )
            )

            return document_collections_response["results"], {
                "total_entries": document_collections_response["total_entries"]
            }

        @self.router.get("/collection/{collection_id}/documents")
        @self.base_endpoint
        async def documents_in_collection_app(
            collection_id: str = Path(..., description="Collection ID"),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedDocumentOverviewResponse,
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get documents in a collection.", 403
                )
            collection_uuid = UUID(collection_id)
            documents_in_collection_response = (
                await self.service.documents_in_collection(
                    collection_uuid, offset, limit
                )
            )

            return documents_in_collection_response["results"], {
                "total_entries": documents_in_collection_response[
                    "total_entries"
                ]
            }



from uuid import UUID

from pydantic import BaseModel, Field

from shared.api.models.base import ResultsWrapper


class KGCreationResponse(BaseModel):
    message: str = Field(
        ...,
        description="A message describing the result of the restructure request.",
    )
    task_id: UUID = Field(
        ...,
        description="The task ID of the restructure request.",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "message": "Graph creation queued successfully.",
                "task_id": "c68dc72e-fc23-5452-8f49-d7bd46088a96",
            }
        }


class KGEnrichmentResponse(BaseModel):
    message: str = Field(
        ...,
        description="A message describing the result of the restructure request.",
    )
    task_id: UUID = Field(
        ...,
        description="The task ID of the restructure request.",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "message": "Graph enrichment queued successfuly.",
                "task_id": "c68dc72e-fc23-5452-8f49-d7bd46088a96",
            }
        }


WrappedKGCreationResponse = ResultsWrapper[KGCreationResponse]
WrappedKGEnrichmentResponse = ResultsWrapper[KGEnrichmentResponse]



from pathlib import Path
from typing import Optional

import yaml
from fastapi import Body, Depends
from fastapi.responses import StreamingResponse

from core.base import (
    GenerationConfig,
    KGSearchSettings,
    Message,
    R2RException,
    RunType,
    VectorSearchSettings,
)
from core.base.api.models import (
    WrappedRAGAgentResponse,
    WrappedRAGResponse,
    WrappedSearchResponse,
)
from core.base.providers import OrchestrationProvider

from ..services.retrieval_service import RetrievalService
from .base_router import BaseRouter


class RetrievalRouter(BaseRouter):
    def __init__(
        self,
        service: RetrievalService,
        run_type: RunType = RunType.RETRIEVAL,
        orchestration_provider: Optional[OrchestrationProvider] = None,
    ):
        super().__init__(service, run_type, orchestration_provider)
        self.service: RetrievalService = service  # for type hinting

    def _load_openapi_extras(self):
        yaml_path = (
            Path(__file__).parent / "data" / "retrieval_router_openapi.yml"
        )
        with open(yaml_path, "r") as yaml_file:
            yaml_content = yaml.safe_load(yaml_file)
        return yaml_content

    def _register_workflows(self):
        pass

    def _setup_routes(self):
        search_extras = self.openapi_extras.get("search", {})
        search_descriptions = search_extras.get("input_descriptions", {})

        @self.router.post(
            "/search",
            openapi_extra=search_extras.get("openapi_extra"),
        )
        @self.base_endpoint
        async def search_app(
            query: str = Body(
                ..., description=search_descriptions.get("query")
            ),
            vector_search_settings: VectorSearchSettings = Body(
                default_factory=VectorSearchSettings,
                description=search_descriptions.get("vector_search_settings"),
            ),
            kg_search_settings: KGSearchSettings = Body(
                default_factory=KGSearchSettings,
                description=search_descriptions.get("kg_search_settings"),
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedSearchResponse,
        ):
            print("kg_search_settings = ", kg_search_settings)
            """
            Perform a search query on the vector database and knowledge graph.

            This endpoint allows for complex filtering of search results using PostgreSQL-based queries.
            Filters can be applied to various fields such as document_id, and internal metadata values.


            Allowed operators include `eq`, `neq`, `gt`, `gte`, `lt`, `lte`, `like`, `ilike`, `in`, and `nin`.

            """
            user_collections = set(auth_user.collection_ids)
            selected_collections = set(
                vector_search_settings.selected_collection_ids
            )
            allowed_collections = user_collections.intersection(
                selected_collections
            )
            if selected_collections - allowed_collections != set():
                raise ValueError(
                    "User does not have access to the specified collection(s): "
                    f"{selected_collections - allowed_collections}"
                )

            filters = {
                "$or": [
                    {"user_id": {"$eq": str(auth_user.id)}},
                    {
                        "collection_ids": {
                            "$overlap": list(allowed_collections)
                        }
                    },
                ]
            }
            if vector_search_settings.filters != {}:
                filters = {"$and": [filters, vector_search_settings.filters]}

            vector_search_settings.filters = filters
            results = await self.service.search(
                query=query,
                vector_search_settings=vector_search_settings,
                kg_search_settings=kg_search_settings,
            )
            return results

        rag_extras = self.openapi_extras.get("rag", {})
        rag_descriptions = rag_extras.get("input_descriptions", {})

        @self.router.post(
            "/rag",
            openapi_extra=rag_extras.get("openapi_extra"),
        )
        @self.base_endpoint
        async def rag_app(
            query: str = Body(..., description=rag_descriptions.get("query")),
            vector_search_settings: VectorSearchSettings = Body(
                default_factory=VectorSearchSettings,
                description=rag_descriptions.get("vector_search_settings"),
            ),
            kg_search_settings: KGSearchSettings = Body(
                default_factory=KGSearchSettings,
                description=rag_descriptions.get("kg_search_settings"),
            ),
            rag_generation_config: GenerationConfig = Body(
                default_factory=GenerationConfig,
                description=rag_descriptions.get("rag_generation_config"),
            ),
            task_prompt_override: Optional[str] = Body(
                None, description=rag_descriptions.get("task_prompt_override")
            ),
            include_title_if_available: bool = Body(
                False,
                description=rag_descriptions.get("include_title_if_available"),
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedRAGResponse,
        ):
            """
            Execute a RAG (Retrieval-Augmented Generation) query.

            This endpoint combines search results with language model generation.
            It supports the same filtering capabilities as the search endpoint,
            allowing for precise control over the retrieved context.

            The generation process can be customized using the rag_generation_config parameter.
            """
            print("kg_search_settings = ", kg_search_settings)

            allowed_collections = set(auth_user.collection_ids)
            filters = {
                "$or": [
                    {"user_id": str(auth_user.id)},
                    {
                        "collection_ids": {
                            "$overlap": list(allowed_collections)
                        }
                    },
                ]
            }
            if vector_search_settings.filters != {}:
                filters = {"$and": [filters, vector_search_settings.filters]}

            vector_search_settings.filters = filters

            response = await self.service.rag(
                query=query,
                vector_search_settings=vector_search_settings,
                kg_search_settings=kg_search_settings,
                rag_generation_config=rag_generation_config,
                task_prompt_override=task_prompt_override,
                include_title_if_available=include_title_if_available,
            )

            if rag_generation_config.stream:

                async def stream_generator():
                    async for chunk in response:
                        yield chunk

                return StreamingResponse(
                    stream_generator(), media_type="application/json"
                )
            else:
                return response

        agent_extras = self.openapi_extras.get("agent", {})
        agent_descriptions = agent_extras.get("input_descriptions", {})

        @self.router.post(
            "/agent",
            openapi_extra=agent_extras.get("openapi_extra"),
        )
        @self.base_endpoint
        async def agent_app(
            messages: list[Message] = Body(
                ..., description=agent_descriptions.get("messages")
            ),
            vector_search_settings: VectorSearchSettings = Body(
                default_factory=VectorSearchSettings,
                description=agent_descriptions.get("vector_search_settings"),
            ),
            kg_search_settings: KGSearchSettings = Body(
                default_factory=KGSearchSettings,
                description=agent_descriptions.get("kg_search_settings"),
            ),
            rag_generation_config: GenerationConfig = Body(
                default_factory=GenerationConfig,
                description=agent_descriptions.get("rag_generation_config"),
            ),
            task_prompt_override: Optional[str] = Body(
                None,
                description=agent_descriptions.get("task_prompt_override"),
            ),
            include_title_if_available: bool = Body(
                True,
                description=agent_descriptions.get(
                    "include_title_if_available"
                ),
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
            response_model=WrappedRAGAgentResponse,
        ):
            """
            Implement an agent-based interaction for complex query processing.

            This endpoint supports multi-turn conversations and can handle complex queries
            by breaking them down into sub-tasks. It uses the same filtering capabilities
            as the search and RAG endpoints for retrieving relevant information.

            The agent's behavior can be customized using the rag_generation_config and
            task_prompt_override parameters.
            """
            # TODO - Don't just copy paste the same code, refactor this
            user = auth_user
            allowed_collections = set(user.collection_ids)
            filters = {
                "$or": [
                    {"user_id": str(user.id)},
                    {
                        "collection_ids": {
                            "$overlap": list(allowed_collections)
                        }
                    },
                ]
            }
            if vector_search_settings.filters != {}:
                filters = {"$and": [filters, vector_search_settings.filters]}

            vector_search_settings.filters = filters

            try:
                response = await self.service.agent(
                    messages=messages,
                    vector_search_settings=vector_search_settings,
                    kg_search_settings=kg_search_settings,
                    rag_generation_config=rag_generation_config,
                    task_prompt_override=task_prompt_override,
                    include_title_if_available=include_title_if_available,
                )

                if rag_generation_config.stream:

                    async def stream_generator():
                        async for chunk in response:
                            yield chunk

                    return StreamingResponse(
                        stream_generator(), media_type="application/json"
                    )
                else:
                    return {"messages": response}
            except Exception as e:
                raise R2RException(str(e), 500)
